/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#define COLOR_MAX 6
#define COLOR_SUFFIX_MAX 2
#define COLOR_NAMES {"RGB24","PAL8","RGB16","YUV422","YUV420","GREY8","RGB15"}
#define COLOR_BITS {24,8,16,16,12,8,16}
#define COLOR_PAL_NUM {3,1,3,3,3,1,3}
#define COLOR_PAL_BITS {{8,8,8},{8,0,0},{8,8,8},{8,4,4},{8,2,2},{8,0,0},{8,8,8}}
#define COLOR_PACKED_NUM {1,1,1,1,2,1,1}
#define COLOR_PACKED_BITS {{24,0},{8,0},{16,0},{16,0},{16,8},{8,0},{16,0}}
#define COLOR_ALIGN {{1,1},{1,1},{1,1},{2,1},{2,2},{1,1},{1,1}}
#define COLOR_SUFFIX "PS"
#define COLOR_VALIDATE(x) ASSERT((x) >= 0 && (x) < (COLOR_MAX+1)*(COLOR_SUFFIX_MAX+1),(INTMSG("%i is not a valid color type"),(x)))
typedef int color_t;

#define color_num(x) (x/(COLOR_SUFFIX_MAX+1))
#define color_suf(x) (x%(COLOR_SUFFIX_MAX+1))

static inline color_t color_val(const char *name)
{
	const char *names[] = COLOR_NAMES;
	const char *suffix = COLOR_SUFFIX;
	int n, s, l;
	l = strlen(name);
	for (s = 1; s <= COLOR_SUFFIX_MAX; s++) {
		if (name[l-1] == suffix[s-1]) {
			l--;
		}
	}
	if (s > COLOR_SUFFIX_MAX) s = 0;
	for (n = 0; n <= COLOR_MAX; n++) {
		if (!strncmp(name, names[n], l)) return n*(COLOR_SUFFIX_MAX+1)+s;
	}
	ERROR(INTMSG("'%s' is not a valid color type"),name);
	END(1);
	return 0;
}

static const inline char *color_name(color_t num)
{
	static char buf[10] = "";
	const char *names[] = COLOR_NAMES;
	const char *suffix = COLOR_SUFFIX;
	COLOR_VALIDATE(num);
	strcpy(buf,names[color_num(num)]);
	buf[strlen(buf)+1] = 0;
	if (color_suf(num)) buf[strlen(buf)] = suffix[color_suf(num)-1];
	return buf;
}

static inline void color_check(color_t num, int width, int height, const char *name, int xalign, int yalign)
{
	const int align[][2] = COLOR_ALIGN;
	COLOR_VALIDATE(num);
	if (xalign) ASSERT(!(width%xalign),(MSG("width must be multiple of %i for %s"),xalign,name));
	else ASSERT(!(width%align[color_num(num)][0]),(MSG("width must be multiple of %i for %s"),align[color_num(num)][0],name));
	if (yalign) ASSERT(!(height%yalign),(MSG("height must be multiple of %i for %s"),yalign,name));
	else ASSERT(!(height%align[color_num(num)][1]),(MSG("height must be multiple of %i for %s"),align[color_num(num)][1],name));
}

static inline int color_bytes(color_t num, int width, int height)
{
	const int bits[] = COLOR_BITS;
	COLOR_VALIDATE(num);
	return (width*height*bits[color_num(num)])/8;
}

static inline unsigned char **color_mkptrs(color_t num, int width, int height, unsigned char *base)
{
	unsigned char **ptr = NULL;
	int i;
	COLOR_VALIDATE(num);

	if (color_suf(num) == 1) { // palettized
		const int comp_num[] = COLOR_PAL_NUM;
		const int comp_bits[][3] = COLOR_PAL_BITS;
		SAFE_MALLOC(ptr,comp_num[color_num(num)]*sizeof(unsigned char *));
		for (i = 0; i < comp_num[color_num(num)]; i++) base = ptr[i] = base+((comp_bits[color_num(num)][i]*width*height)/8);
	} else {
		const int comp_num[] = COLOR_PACKED_NUM;
		const int comp_bits[][2] = COLOR_PACKED_BITS;
		SAFE_MALLOC(ptr,height*sizeof(unsigned char *));
		for (i = 0; i < height; i++) base = ptr[i] = base+((comp_bits[color_num(num)][i%comp_num[color_num(num)]]*width)/8);
	}
	return ptr;
}

#define color_freeptrs(ptr) SAFE_FREE(ptr)
