/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <pthread.h>
#ifdef HAVE_SEMAPHORE_H
#include <semaphore.h>
#endif
#include <setjmp.h>
#include "version.h"
#include "output.h"
#include "grab.h"
#include "sound.h"
#include "read_sound.h"
#define ERROR_DEBUG
#include "error.h"
#include "mpeg.h"
#include "read.h"
#include "main.h"
#include "convert.h"
#include "transport.h"
#include "configuration.h"
#define SYMBOL(x) main_ ## x
#ifdef __CYGWIN__
// these are all kludges, because cygwin doesn't yet have complete multithreading
#define PTH_T(x) &x
#define pthread_exit(x) my_pthread_exit(x)
#define sem_init(x,y,z) my_sem_init(x,y,z)
#define sem_wait(x) my_sem_wait(x)
#define sem_trywait(x) my_sem_trywait(x)
#define sem_post(x) my_sem_post(x)
#define sem_destroy(x) my_sem_destroy(x)
#define sem_t void *
#else
#define PTH_T(x) x
#endif

/************ Prototypes ************/

void libgrab_control(int action);
void libgrab_configure(const char *option, const char *value);
void libgrab_inquire(const char *option, char value[PATH_MAX]);
int libgrab_getprogress(char *src, char *dest, char *conv);
void main_endprogress();
void main_setprogressresult(int end);
void main_setprogresscnt(int cnt);
void main_sendprogress();
void main_setprogressdest(const char *dest);
void main_setprogresssrc(const char *src);
void main_setprogressconv(const char *conv);
void main_mpeg_remove(int n);
static inline void setup_sound();
static void *audio_loop(void *ign);
static inline void cleanup_sound();
static inline void setup_grab();
static inline void cleanup_grab();
static inline int control_event(int wait);
static inline void execute_video_grab();
static inline void setup_convert();
static inline void cleanup_convert();
static inline int execute_video_convert(int n);
static void work_loop();
void libgrab_init();
void *main_init(void *ign);
void main_end();
void main_signal_handler(int n);
void main_exit_handler();
void main_signal_abort(int n);
void error_abort(int n);
void my_pthread_exit(void *ign);
void my_sem_init(sem_t *x, int y, int z);
void my_sem_wait(sem_t *x);
int my_sem_trywait(sem_t *x);
void my_sem_post(sem_t *x);
void my_sem_destroy(sem_t *x);

/************************************/

void main_inquire(const char *, char *);

char error_text[PATH_MAX] = "";

sigjmp_buf main_abort;
int libgrab_progresspipe = -1, main_outpipe = -1, libgrab_errorpipe = -1, main_erroutpipe = -1;
int main_aborting = 0;
int main_exiting = 0;
int error_ignore = 0;
int main_soundframe = INT_MAX;
int main_videoframe = INT_MAX;
int main_sound_running = 0;
int main_retval = 1;
pthread_t main_thread, main_parent, main_sound_thread;
sigset_t main_sigmask;

#define C_CONFIGURE 'C'
#define C_INQUIRE 'I'

struct {
	sem_t new_data;
	sem_t data_read;
	int action;
	char option[64];
	char value[PATH_MAX];
} main_control_msg;

struct {
	sem_t data_read;
	char src[PATH_MAX];
	char dest[PATH_MAX];
	char conv[PATH_MAX];
	int cnt;
} main_progress_msg;

void libgrab_control(int action)
{
#ifdef NOTHREAD
	main_start();
#else
	sem_wait(&main_control_msg.data_read);
	ASSERT(action == C_START || action == C_STOP || action == C_PAUSE || action == C_CONT,(INTMSG("unknown control code")));
	main_control_msg.action = action;
	sem_post(&main_control_msg.new_data);
#endif
}

void libgrab_configure(const char *option, const char *value)
{
#ifdef NOTHREAD
	main_configure(option,value);
#else
	sem_wait(&main_control_msg.data_read);
	ASSERT(strlen(value) <= PATH_MAX,(INTMSG("configure value too long '%s' (option %s)"),value,option));
	ASSERT(strlen(option) <= 64,(INTMSG("configure option name too long")));
	main_control_msg.action = C_CONFIGURE;
	strcpy(main_control_msg.option,option);
	strcpy(main_control_msg.value,value);
	sem_post(&main_control_msg.new_data);
#endif
}

void libgrab_inquire(const char *option, char value[PATH_MAX])
{
#ifdef NOTHREAD
	main_inquire(option,value);
#else
	sem_wait(&main_control_msg.data_read);
	ASSERT(strlen(option) <= 64,(INTMSG("configure option name too long")));
	main_control_msg.action = C_INQUIRE;
	strcpy(main_control_msg.option,option);
	sem_post(&main_control_msg.new_data);
	sem_wait(&main_control_msg.data_read);
	strcpy(value,main_control_msg.value);
	sem_post(&main_control_msg.data_read);
#endif
}

int libgrab_getprogress(char *src, char *dest, char *conv)
{
#ifndef NOTHREAD
	int cnt;
	char c;

	SAFE_READ(libgrab_progresspipe,&c,1,"libgrab_progresspipe"); 

	strcpy(src,main_progress_msg.src);
	strcpy(dest,main_progress_msg.dest);
	strcpy(conv,main_progress_msg.conv);
	cnt = main_progress_msg.cnt;

	sem_post(&main_progress_msg.data_read);

	return cnt;
#endif
}

void main_endprogress()
{
	main_setprogresscnt(-main_progress_msg.cnt);
	main_sendprogress();
}

void main_setprogressresult(int end)
{
	*main_progress_msg.src = 0;
	*main_progress_msg.dest = 0;
	*main_progress_msg.conv = 0;
	main_progress_msg.cnt = end;
	main_sendprogress();
}

void main_setprogresscnt(int cnt)
{
	main_progress_msg.cnt = cnt;
}

void main_sendprogress()
{
#ifndef NOTHREAD
	char c = 0;
	SAFE_WRITE(main_outpipe,&c,1,"pipe");
	sem_wait(&main_progress_msg.data_read);
#endif
}

void main_setprogressdest(const char *dest)
{
	ASSERT(!dest || strlen(dest) < PATH_MAX,(INTMSG("destprogress message too long")));
	strcpy(main_progress_msg.dest,dest);
}

void main_setprogresssrc(const char *src)
{
	ASSERT(!src || strlen(src) < PATH_MAX,(INTMSG("srcprogress message too long")));
	strcpy(main_progress_msg.src,src);
}

void main_setprogressconv(const char *conv)
{
	ASSERT(!conv || strlen(conv) < PATH_MAX,(INTMSG("convprogress message too long")));
	strcpy(main_progress_msg.conv,conv);
}


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
int main_do_sound = 1;
int main_sleeptime = 0;
int main_limit = 0;
int main_width = 640;
int main_height = 480;
int main_do_grab = 1;
int main_do_convert = 0;
int main_mpeg = 0;
int main_lowdisk = 0;
int main_always_abort = 1;
char main_grablog[PATH_MAX] = "";
char main_convfmt[PATH_MAX] = "";
char main_convname[PATH_MAX] = "./";
char main_conv_transport[PATH_MAX] = "file";
int main_first_conv = 0;
char main_tmpname[PATH_MAX] = "";
int main_msleep;
int main_fbs = 32;

START_OPTIONS
OPTREDIRECT(grabdevice,grab,device);
OPTREDIRECT(channel,grab,channel);
OPTREDIRECT(norm,grab,norm);
OPTREDIRECT(channel list,grab,channel list);
INTOPTION(do_sound,0,1) FORWARD(mpeg,sound);
INTOPTION(msleep,0,INT_MAX);
INTOPTION(fbs,2,32);
INTOPTION(sleeptime,0,INT_MAX);
OPTREDIRECT(skip,grab,skip);
OPTREDIRECT(grabinfo,grab,info);
INTOPTION(limit,0,INT_MAX);
INTOPTION(width,32,768);
INTOPTION(height,32,768);
INTOPTION(first_conv,0,INT_MAX);
OPTREDIRECT(video_bitrate,mpeg,bitrate);
OPTREDIRECT(rate,transport,rate);
OPTREDIRECT(srate,transport,srate);
OPTREDIRECT(audio_bitrate,mpeg,audio_rate);
INTOPTION(lowdisk,0,1) FORWARD(mpeg,lowdisk);
OPTREDIRECT(conv_log,mpeg,logname);
OPTREDIRECT(inputname,transport,rname);
OPTREDIRECT(output_data,output,data);
OPTREDIRECT(grabname,transport,wname);
OPTREDIRECT(osoundname,transport,swname);
OPTREDIRECT(isoundname,transport,srname);
STROPTION(convname);
INTOPTION(always_abort,0,1);
STROPTION(grablog); if (set) FORWARD(grab,log);
OPTREDIRECT(osounddata,sound,data);
OPTREDIRECT(isounddata,read_sound,data);
OPTREDIRECT(do_sync,transport,sync);
OPTREDIRECT(input_transport,transport,rformat str);
OPTREDIRECT(grab_transport,transport,wformat str);
STROPTION(conv_transport);
OPTION(transport list) NOSET O_GET {
	char tmp[PATH_MAX];
	int i;
	transport_inquire("format nr",tmp);
	i = atoi(tmp);
	value[0] = 0;
	while(i-->0) {
		sprintf(tmp,"%i",i);
		transport_configure("wformat",tmp);
		transport_inquire("wformat str",tmp);
		strcat(value," ");
		strcat(value,tmp);
	}
}
OPTREDIRECT(input_sound_fmt,read_sound,file_fmt str);
OPTREDIRECT(output_sound_fmt,sound,file_fmt str);
OPTION(input_file_fmt) O_SET {
	if (strchr(value,'-')) *(strchr(value,'-')) = 0;
	FORWARD(read,file_fmt str);
	main_do_grab = 0;
} O_GET {
	if (main_do_grab) value[0] = 0;
}
OPTION(grab_file_fmt) O_SET {
	if (value[0] == 0) {
		main_do_grab = 0;
	} else {
		FORWARD(output,file_fmt str);
		main_do_grab = 1;
	}
} O_GET {
	FORWARD(output,file_fmt str);
	if (!main_do_grab) value[0] = 0;
}
OPTION(conv_file_fmt) O_SET {
	if (value[0] == 0) {
		main_do_convert = 0;
	} else {
		main_do_convert = 1;
		if (!strcmp(value,"mpeg-1")) {
			strcpy(main_convfmt,"yuv");
			mpeg_configure("type","1");
			main_mpeg = 1;
		} else if (!strcmp(value,"mpeg-2")) {
			strcpy(main_convfmt,"yuv");
			mpeg_configure("type","2");
			main_mpeg = 1;
		} else {
			strcpy(main_convfmt,value);
			main_mpeg = 0;
		}
	}
} O_GET {
	if (!main_do_convert) {
		value[0] = 0;
	} else {
		if (main_mpeg) {
			char w[2];
			mpeg_inquire("type",w);
			strcpy(value,"mpeg-");
			strcat(value,w);
		} else {
			strcpy(value,main_convfmt);
		}
	}
}
OPTION(output_sound_fmt list) NOSET O_GET {
	char tmp[PATH_MAX];
	int i;
	sound_inquire("file_fmt nr",tmp);
	i = atoi(tmp);
	value[0] = 0;
	while(i-->0) {
		sprintf(tmp,"%i",i);
		sound_configure("file_fmt",tmp);
		sound_inquire("file_fmt str",tmp);
		strcat(value," ");
		strcat(value,tmp);
	}
}
OPTION(grab_file_fmt list) NOSET O_GET {
	char tmp[PATH_MAX];
	int i;
	output_inquire("file_fmt nr",tmp);
	i = atoi(tmp);
	value[0] = 0;
	while(i-->0) {
		sprintf(tmp,"%i",i);
		output_configure("file_fmt",tmp);
		output_inquire("file_fmt str",tmp);
		strcat(value," ");
		strcat(value,tmp);
	}
}
OPTION(conv_file_fmt list) NOSET O_GET {
	main_inquire("grab_file_fmt list",value);
	strcat(value," mpeg-1 mpeg-2");
}
OPTION(input_sound_fmt list) NOSET O_GET {
	char tmp[PATH_MAX];
	int i;
	read_sound_inquire("file_fmt nr",tmp);
	i = atoi(tmp);
	value[0] = 0;
	while(i-->0) {
		sprintf(tmp,"%i",i);
		read_sound_configure("file_fmt",tmp);
		read_sound_inquire("file_fmt str",tmp);
		strcat(value," ");
		strcat(value,tmp);
	}
}
OPTION(input_file_fmt list) NOSET O_GET {
	char tmp[PATH_MAX];
	int i;
	read_inquire("file_fmt nr",tmp);
	i = atoi(tmp);
	value[0] = 0;
	while(i-->0) {
		sprintf(tmp,"%i",i);
		read_configure("file_fmt",tmp);
		read_inquire("file_fmt str",tmp);
		strcat(value," ");
		strcat(value,tmp);
	}
}
OPTREDIRECT(outputdata,output,data);
OPTREDIRECT(inputdata,read,data);
END_OPTIONS

void main_mpeg_remove(int n)
{
	transport_wremove(n);
}

#define forward_data(in,out,val) do { strcpy(tmp,#val); in ## _inquire("data",tmp); sprintf(tmp2,#val "=%s",tmp); out ## _configure("data",tmp2); } while (0)
#define xforward(in,out,ival,oval) do { in ## _inquire(#ival,tmp); out ## _configure(#oval,tmp); } while (0)
#define forward(in,out,val) do { in ## _inquire(#val,tmp); out ## _configure(#val,tmp); } while (0)

static inline void setup_sound()
{
	
	if (main_do_sound) {
		pthread_attr_t pattr;

		grab_signal = read_sound_setnext;

		main_soundframe = 0;
		pthread_attr_init(&pattr);
#ifdef PTHREAD_CREATE_DETACHED
		pthread_attr_setdetachstate(&pattr,PTHREAD_CREATE_DETACHED);
#endif
		EQNULL((void *)pthread_create(&main_sound_thread,&pattr,audio_loop,NULL),(ERRMSG("pthread_create")));
		pthread_attr_destroy(&pattr);

	} else {
		main_soundframe = INT_MAX;
	}

}

static void *audio_loop(void *ign)
{
	char tmp[PATH_MAX], tmp2[PATH_MAX];
	int soundfb, len;
	int soundpicnr = 0;

	read_sound_start();
	forward_data(read_sound,sound,bits);
	forward_data(read_sound,sound,rate);
	forward_data(read_sound,sound,channels);
	forward(read_sound,sound,fbs);
	forward(read_sound,sound,buf);
	sound_start();

	main_sound_running = 1;
	while (main_sound_running) {
		len = read_sound_get(&soundfb,&soundpicnr);
		while (soundpicnr-50 > main_videoframe) usleep(10000);
		sound_write(soundfb, soundpicnr, len);
	
		main_soundframe = soundpicnr;
		soundpicnr++;
	}
	sound_stop();
	read_sound_stop();

	main_sound_running = 1;
	return NULL;
}

static inline void cleanup_sound()
{
	int i;

	if (main_do_sound) {
		main_sound_running = 0;
		main_videoframe = INT_MAX;
		for (i = 0; i < 50; i++) {
			if (main_sound_running) break;
			usleep(40000);
		}
		if (i == 50) pthread_kill(PTH_T(main_sound_thread),SIGABRT);
	}
}


static inline void setup_grab()
{
	char tmp[PATH_MAX];

	sprintf(tmp,"%i",main_width);
	grab_configure("width",tmp);
	read_configure("width",tmp);
	mpeg_configure("width",tmp);
	convert_configure("inwidth",tmp);
	convert_configure("outwidth",tmp);
	sprintf(tmp,"width=%i",main_width);
	output_configure("data",tmp);

	sprintf(tmp,"%i",main_height);
	grab_configure("height",tmp);
	read_configure("height",tmp);
	mpeg_configure("height",tmp);
	convert_configure("inheight",tmp);
	convert_configure("outheight",tmp);
	sprintf(tmp,"height=%i",main_height);
	output_configure("data",tmp);

	if (main_sleeptime) grab_configure("sleep","1");
	else grab_configure("sleep","0");
	grab_inquire("norm",tmp);
	if (!strcmp(tmp,"NTSC")) {
		mpeg_configure("ntsc","1");
//		output_configure("data","framerate=30");
	} else {
		mpeg_configure("ntsc","0");
//		output_configure("data","framerate=25");
	}

	strcpy(tmp,"color_fmt");
	xforward(output,convert,data,outfmt);
	xforward(convert,grab,outfmt,color_fmt);
	xforward(grab,convert,color_fmt,infmt);
	sprintf(tmp,"%i",main_fbs);
	grab_configure("fbs",tmp);

	setup_sound();
	grab_start();

	forward(grab,convert,fbs);
	xforward(grab,convert,buf,inbuf);

	convert_start();

	forward(convert,output,fbs);
	xforward(convert,output,outbuf,buf);

	output_start();
		sleep(1);
}

static inline void cleanup_grab()
{
	cleanup_sound();
	grab_stop();
	output_stop();
	convert_stop();
}

static inline int control_event(int wait)
{
	int rc = 0;

#ifndef NOTHREAD
	if (wait) sem_wait(&main_control_msg.new_data);
	else rc = sem_trywait(&main_control_msg.new_data);

	if (!rc) {
		switch (main_control_msg.action) {
		case C_CONFIGURE:
		case C_INQUIRE:
			ERROR(MSG("no configure/inquire messages while running"));
			END(-1);
			return 1;
		case C_STOP:
			sem_post(&main_control_msg.data_read);
			return 1;
		case C_START:
			ERROR(INTMSG("already running!"));
			END(-1);
			return 1;
		case C_CONT:
			sem_post(&main_control_msg.data_read);
			return 0;
		case C_PAUSE:
			sem_post(&main_control_msg.data_read);
			return control_event(1);
		default:
			ERROR(INTMSG("unknown control message '%c'"),main_control_msg.action);
			END(-1);
			return 1;
		}
	}
#endif
	return 0;
}

static inline void execute_video_grab()
{
	int grabfb, grabpicnr;

	grabfb = grab_get();
	grabpicnr = grab_get_picnr();
	grabfb = convert_convert(grabfb);
	output_write(grabfb,grabpicnr);
	main_videoframe = grabpicnr;

	main_sendprogress();
	if (main_sleeptime) sleep(main_sleeptime);
	grab_release();
}

static inline void setup_convert()
{
	char tmp[PATH_MAX], tmp2[PATH_MAX];

	if (main_do_grab) {
		output_inquire("file_fmt str",tmp);
		if (strchr(tmp,'-')) *(strchr(tmp,'-')) = 0;
		read_configure("file_fmt str",tmp);
		transport_inquire("wname",main_tmpname);
		transport_configure("rname",main_tmpname);
		transport_inquire("wformat str",tmp);
		transport_configure("rformat str",tmp);
	}

	if (main_limit > 0){
		sprintf(tmp,"%i",main_limit-1);
		mpeg_configure("last",tmp);
	}
	sprintf(tmp,"%i",main_first_conv);
	mpeg_configure("first",tmp);

	output_configure("file_fmt str",main_convfmt);
	transport_configure("wformat str",main_conv_transport);

	read_start();

	read_inquire("fbs",tmp2);
	convert_configure("fbs",tmp2);
	read_inquire("buf",tmp2);
	convert_configure("inbuf",tmp2);

	read_inquire("width",tmp2);
	main_width = atoi(tmp2);
	read_inquire("height",tmp2);
	main_height = atoi(tmp2);

	sprintf(tmp,"%i",main_width);
	grab_configure("width",tmp);
	read_configure("width",tmp);
	mpeg_configure("width",tmp);
	convert_configure("inwidth",tmp);
	convert_configure("outwidth",tmp);
	sprintf(tmp,"width=%i",main_width);
	output_configure("data",tmp);

	sprintf(tmp,"%i",main_height);
	grab_configure("height",tmp);
	read_configure("height",tmp);
	mpeg_configure("height",tmp);
	convert_configure("inheight",tmp);
	convert_configure("outheight",tmp);
	sprintf(tmp,"height=%i",main_height);
	output_configure("data",tmp);

	read_inquire("color_fmt",tmp2);
	convert_configure("infmt",tmp2);
	strcpy(tmp2,"color_fmt");
	output_inquire("data",tmp2);
	convert_configure("outfmt",tmp2);

	convert_start();

	convert_inquire("fbs",tmp2);
	output_configure("fbs",tmp2);
	convert_inquire("outbuf",tmp2);
	output_configure("buf",tmp2);

	if (main_mpeg) {
		tmpnam(main_tmpname);
		strcat(main_tmpname,"%04d.yuv");
		mpeg_configure("tmpname",main_tmpname);
		mpeg_configure("trace","1");

		transport_configure("wname",main_tmpname);
	} else {
		transport_configure("wname",main_convname);
		setup_sound();
	}

	output_start();
}

static inline void cleanup_convert()
{
	cleanup_sound();
	read_stop();
	mpeg_stop();
	output_stop();
	convert_stop();
}

static inline int execute_video_convert(int n)
{
	n = read_read(0,n);
	convert_convert(0);
	output_write(0,n); 
	if (main_lowdisk) transport_rremove(n);

	main_setprogresscnt(n);
	main_setprogressconv(main_progress_msg.dest);
	main_sendprogress();

	if (main_msleep) usleep(main_msleep*1000);

	return 1;
}

static inline int execute_video_convert_m(int n)
{
	read_read(0,n);
	convert_convert(0);
	output_write(0,n); 
	if (main_lowdisk) transport_rremove(n);

	main_setprogresscnt(n);
	main_setprogressconv(main_progress_msg.dest);
	main_sendprogress();

	if (main_msleep) usleep(main_msleep*1000);

	return 1;
}

static void work_loop()
{
	volatile int picnr = 0;

	main_progress_msg.cnt = 0;

	if (main_do_grab) {
		if (sigsetjmp(main_abort,1)) { cleanup_grab(); return; }

		setup_grab();
		picnr = 0;

		while (!main_limit || main_limit >= picnr) {
			if (control_event(0)) break;

			main_setprogresscnt(picnr);
			execute_video_grab();

			picnr++;
		}
		cleanup_grab();

	}

	main_setprogressresult(picnr);

	if (main_do_convert) {
		if (sigsetjmp(main_abort,1)) { cleanup_convert(); return; }

		setup_convert();
		picnr = main_first_conv;

		if (main_mpeg) {
			mpeg_start();
		} else {
			while (!main_limit || main_limit >= picnr) {
				if (control_event(0)) break;

				main_setprogresscnt(picnr);
				if (!execute_video_convert(picnr)) break;

				picnr++;
			}
			main_setprogressresult(picnr);
		}
		cleanup_convert();
	}
}

void libgrab_init()
{
	int pipes[2];
	pthread_attr_t pattr;
	struct sigaction sa;
	sigset_t mask;

	if (sigsetjmp(main_abort,1)) {
		exit(main_retval);
	}

	sem_init(&main_control_msg.data_read,0,0);
	sem_init(&main_control_msg.new_data,0,0);
	sem_init(&main_progress_msg.data_read,0,0);

        NOTNEG1(sigfillset(&main_sigmask),(ERRMSG("sigfillset")));
        NOTNEG1(sigdelset(&main_sigmask,SIGUSR1),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&main_sigmask,SIGQUIT),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&main_sigmask,SIGKILL),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&main_sigmask,SIGABRT),(ERRMSG("sigdelset")));

        NOTNEG1(sigfillset(&mask),(ERRMSG("sigfillset")));
        NOTNEG1(sigdelset(&mask,SIGUSR1),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGINT),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGABRT),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGQUIT),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGKILL),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGTERM),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGCHLD),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGHUP),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGSEGV),(ERRMSG("sigdelset")));
        NOTNEG1(sigdelset(&mask,SIGPIPE),(ERRMSG("sigdelset")));

        NOTNEG1(pthread_sigmask(SIG_SETMASK,(void *)&mask,NULL),(ERRMSG("sigprocmask")));

        sa.sa_handler = main_signal_handler;
        sa.sa_flags = SA_NOCLDSTOP;
        NOTNEG1(pthread_sigmask(SIG_SETMASK,NULL,(void *)&(sa.sa_mask)),(ERRMSG("sigprocmask")));
        NOTNEG1(sigaddset(&(sa.sa_mask),SIGUSR1),(ERRMSG("sigdelset")));
        NOTNEG1(sigaction(SIGUSR1,&sa,NULL),(ERRMSG("sigaction")));

        NOTNEG1(sigdelset(&(sa.sa_mask),SIGUSR1),(ERRMSG("sigdelset")));
        sa.sa_handler = main_signal_abort;
        NOTNEG1(sigaction(SIGINT,&sa,NULL),(ERRMSG("sigaction")));
        NOTNEG1(sigaction(SIGTERM,&sa,NULL),(ERRMSG("sigaction")));
	NOTNEG1(sigaction(SIGHUP,&sa,NULL),(ERRMSG("sigaction")));
	NOTNEG1(sigaction(SIGPIPE,&sa,NULL),(ERRMSG("sigaction")));

        sa.sa_handler = (void *)pthread_exit;
	NOTNEG1(sigaction(SIGABRT,&sa,NULL),(ERRMSG("sigaction")));

	main_parent = pthread_self();

	NOTNEG1(pipe(pipes),(ERRMSG("pipe")));
	libgrab_progresspipe = pipes[0];
	main_outpipe = pipes[1];
	fcntl(libgrab_progresspipe,F_SETFL,O_NONBLOCK);
	NOTNEG1(pipe(pipes),(ERRMSG("pipe")));
	libgrab_errorpipe = pipes[0];
	main_erroutpipe = pipes[1];
	fcntl(libgrab_errorpipe,F_SETFL,O_NONBLOCK);

	*main_progress_msg.src = 0;
	*main_progress_msg.dest = 0;
	*main_progress_msg.conv = 0;

	atexit(main_exit_handler);

#ifdef NOTHREAD
	main_init(NULL);
#else
	pthread_attr_init(&pattr);
#ifdef PTHREAD_CREATE_DETACHED
	pthread_attr_setdetachstate(&pattr,PTHREAD_CREATE_DETACHED);
#endif
	EQNULL((void *)pthread_create(&main_thread,&pattr,main_init,NULL),(ERRMSG("pthread_create")));
	pthread_attr_destroy(&pattr);
	sem_wait(&main_control_msg.data_read);
	sem_post(&main_control_msg.data_read);
#endif

}

void *main_init(void *ign)
{
        NOTNEG1(pthread_sigmask(SIG_SETMASK,(void *)&main_sigmask,NULL),(ERRMSG("sigprocmask")));

	main_thread = pthread_self();

	transport_otrace = main_setprogressdest;
	transport_itrace = main_setprogresssrc;
	mpeg_strace_func = main_setprogresssrc;
	mpeg_trace_func = main_setprogressconv;
	mpeg_remove_func = main_mpeg_remove;
	mpeg_conv_func = execute_video_convert_m;

	grab_init();
	convert_init();
	output_init();
	read_init();
	mpeg_init();
	read_sound_init();
	sound_init();

	sem_post(&main_control_msg.data_read);
#ifndef NOTHREAD
main_loop:
	if (sigsetjmp(main_abort,1)) {
		while(!sem_trywait(&main_control_msg.data_read));
		while(!sem_trywait(&main_control_msg.new_data));
		sem_post(&main_control_msg.data_read);
		goto main_loop;
	}
	sem_wait(&main_control_msg.new_data);
	switch (main_control_msg.action) {
	case C_CONFIGURE:
		main_configure(main_control_msg.option,main_control_msg.value);
		sem_post(&main_control_msg.data_read);
		break;
	case C_INQUIRE:
		main_inquire(main_control_msg.option,main_control_msg.value);
		sem_post(&main_control_msg.data_read);
		break;
	case C_STOP:
	case C_PAUSE:
	case C_CONT:
		ERROR(INTMSG("not running"));
		END(-1);
		break;
	case C_START:
		sem_post(&main_control_msg.data_read);
		work_loop();
		main_endprogress();
		break;
	default:
		ERROR(INTMSG("unknown control message '%c'"),main_control_msg.action);
		END(-1);
		break;
	}
	goto main_loop;
#endif
}


void main_end()
{
	error_ignore = 1;
	grab_end();
	convert_end();
	output_end();
	read_end();
	mpeg_end();
	sound_end();
	read_sound_end();

	if (libgrab_progresspipe != -1) {
		SAFE_CLOSE(libgrab_progresspipe,"pipe");
		SAFE_CLOSE(main_outpipe,"pipe");
		libgrab_progresspipe = -1;
		main_outpipe = -1;
	}

	if (libgrab_errorpipe != -1) {
		SAFE_CLOSE(libgrab_errorpipe,"pipe");
		SAFE_CLOSE(main_erroutpipe,"pipe");
		libgrab_errorpipe = -1;
		main_erroutpipe = -1;
	}
}

void main_signal_handler(int n)
{
	if (*error_text) {
		char c = 0;
		write(main_erroutpipe,&c,1);
	}
	NOTNEG1(sigprocmask(SIG_SETMASK,&main_sigmask,NULL),(ERRMSG("sigprocmask")));
	main_retval = n;
//	if (main_always_abort) exit(n);	else
	siglongjmp(main_abort,1);
}

void main_exit_handler()
{
	main_end();
}

void main_signal_abort(int n)
{
	ERROR(MSG("Fatal Signal %i caught"),n);
	exit(n);
}

void error_abort(int n)
{
#ifdef NOTHREAD
	exit(1);
#else
	pthread_kill(PTH_T(main_thread),SIGUSR1);
	sleep(2);
#endif
}

#ifdef __CYGWIN__
#undef ERROR
#include <windows.h>
void my_pthread_exit(void *ign)
{
	pthread_t x = pthread_self();
	pthread_kill(PTH_T(x),SIGKILL);
}
void my_sem_init(sem_t *x, int y, int z)
{
        *x = CreateSemaphore(NULL,z,128,NULL);
}
void my_sem_wait(sem_t *x)
{
	WaitForSingleObject(*x,INFINITE);
}
int my_sem_trywait(sem_t *x)
{
	if (WaitForSingleObject(*x,0) != WAIT_OBJECT_0) return 1;
	return 0;
}
void my_sem_post(sem_t *x)
{
	ReleaseSemaphore(*x,1,NULL);
}
void my_sem_destroy(sem_t *x)
{
	CloseHandle(*x);
}
#endif
