/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#if defined(HAVE_LIBUSER32) && defined(HAVE_LIBGDI32)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <windows.h>

extern pthread_t main_thread;
#include "version.h"
#undef ERROR
#include "transport.h"
#define ERROR_DEBUG
#include "error.h"
#include "configuration.h"
#include "color.h"
#define SYMBOL(x) output_Win32_ ## x
#ifdef __CYGWIN__
#define PTH_T(x) &x
#else
#define PTH_T(x) x
#endif

#define CLASSNAME "bttvgrab " VERSION

/************ Prototypes ************/

DWORD WINAPI output_Win32_messageloop(LPVOID ignored);
LRESULT CALLBACK output_Win32_WinProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
void output_Win32_start(const char *unused);
void output_Win32_stop(void);
int output_Win32_write(unsigned char *buffer, int picnr);
void output_Win32_display(int picnr);
void output_Win32_setsync(int i);

/************************************/

int output_Win32_fmts() { return 1; }
const char *output_Win32_fmt(int nr) { return "Win32"; }

#define BUFFERS 3

int output_Win32_picnr[BUFFERS];
HBITMAP output_Win32_dcbm[BUFFERS];
HBITMAP output_Win32_old;
HDC output_Win32_mem = 0;
unsigned char *output_Win32_bitmap[BUFFERS];
int output_Win32_init = 0;
HDC output_Win32_dc = 0;
BITMAPINFO *output_Win32_displaybitmap = NULL;
HWND output_Win32_displaywindow = 0;
HINSTANCE output_Win32_hInst;
HANDLE output_Win32_hth = 0;
int output_Win32_width = 0;
int output_Win32_height = 0;
int output_Win32_bytes = 0;
int output_Win32_abort = 0;
int output_Win32_audio_sync = 0;
int output_Win32_current = 0;
int output_Win32_last = 0;
int output_Win32_first = 0;
char output_Win32_class[4096] = "";
char output_Win32_window[4096] = "";
int output_Win32_x = CW_USEDEFAULT;
int output_Win32_y = CW_USEDEFAULT;
int output_Win32_child;

#define COLOR "RGB24S"

START_DATA
INTOPTION(width,0,INT_MAX);
INTOPTION(height,0,INT_MAX);
CONSTINFO(color_fmt,COLOR);
STROPTION(class);
STROPTION(window);
INTOPTION(x,0,INT_MAX);
INTOPTION(y,0,INT_MAX);
END_DATA

DWORD WINAPI output_Win32_messageloop(LPVOID ignored)
{
	MSG msg;
	RECT     rect;
	int flags=WS_POPUP;
	HWND parent = NULL;
	int i;

	rect.left = 0;
	rect.top = 0;
        rect.right = output_Win32_width;
        rect.bottom = output_Win32_height;

	output_Win32_child=0;
	if (*output_Win32_class) {
		parent = FindWindow(output_Win32_class,output_Win32_window);
		if (parent) {
			flags=WS_CHILD;
			output_Win32_child=1;
		}
	}

	NOTNULL(output_Win32_displaywindow = CreateWindowEx(0,
							    CLASSNAME,
							    CLASSNAME,
							    flags|WS_VISIBLE,
							    output_Win32_x,
							    output_Win32_y,
							    rect.right,
							    rect.bottom,
							    parent,
							    NULL,
							    output_Win32_hInst,
							    NULL),
		(MSG("Could not create main window")));

	output_Win32_displaybitmap = calloc(01, sizeof(BITMAPINFO));
	output_Win32_displaybitmap->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	output_Win32_displaybitmap->bmiHeader.biWidth = output_Win32_width;
	output_Win32_displaybitmap->bmiHeader.biHeight = -output_Win32_height;
	output_Win32_displaybitmap->bmiHeader.biPlanes = 1;
	output_Win32_displaybitmap->bmiHeader.biBitCount = 24;
	output_Win32_displaybitmap->bmiHeader.biCompression = BI_RGB;
	output_Win32_displaybitmap->bmiHeader.biClrUsed = 0;
	output_Win32_displaybitmap->bmiHeader.biSizeImage = output_Win32_bytes;
	output_Win32_displaybitmap->bmiHeader.biClrImportant = 0;

        NOTNULL(output_Win32_dc = GetDC(output_Win32_displaywindow),(MSG("could not initialize display area (DC)")));

	NOTNULL(output_Win32_mem = CreateCompatibleDC(output_Win32_dc),
		(MSG("could not initialize display area (mem)")));
	for (i = 0; i < BUFFERS; i++) {
		NOTNULL(output_Win32_dcbm[i] = CreateDIBSection(output_Win32_mem,
							     output_Win32_displaybitmap,
							     DIB_RGB_COLORS,
							     (void *)&output_Win32_bitmap[i],
							     NULL,
							     0),
			(MSG("could not initialize display area (DIB)")));
		output_Win32_picnr[i] = -1;
	}
	NOTNULL(output_Win32_old = SelectObject(output_Win32_mem, output_Win32_dcbm[0]),
		(MSG("could not initialize display area (Select)")));
	output_Win32_init = 1;

        while (GetMessage(&msg,output_Win32_displaywindow,0,0)) if (!TranslateMessage(&msg)) DispatchMessage(&msg);
	ExitProcess(0);
	return 0;
}

static int audiopic = -1;
LRESULT CALLBACK output_Win32_WinProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HDC         hdc;
	MINMAXINFO  *pInfo;
	PAINTSTRUCT ps;
	RECT rect;
	static int sx = INT_MIN, sy = 0;
	int x, y;
	int cur, i;
	extern void sound_dsp_restart();
	extern void sound_dsp_restop();

        switch (message) {
	case WM_GETMINMAXINFO:
		pInfo = (MINMAXINFO *) lParam;
                pInfo->ptMaxSize.x = output_Win32_width;
                pInfo->ptMaxSize.y = output_Win32_height;
                pInfo->ptMinTrackSize.x = output_Win32_width;
                pInfo->ptMinTrackSize.y = output_Win32_height;
                pInfo->ptMaxTrackSize.x = output_Win32_width;
                pInfo->ptMaxTrackSize.y = output_Win32_height;
                return 0;

	case WM_ACTIVATE:
		if (output_Win32_audio_sync) {
			if (LOWORD(wParam) == WA_INACTIVE) sound_dsp_restop();
			else sound_dsp_restart();
		}
		break;

	case WM_LBUTTONDOWN:
		if (output_Win32_child) break;
		if (wParam != MK_LBUTTON) break;
		sx = LOWORD(lParam);
		sy = HIWORD(lParam);
		return 0;

	case WM_MOUSEMOVE:
		if (output_Win32_child) break;
		if (sx == INT_MIN) break;
		GetWindowRect(output_Win32_displaywindow,&rect);
		x = LOWORD(lParam)-sx+rect.left;
		y = HIWORD(lParam)-sy+rect.top;
		MoveWindow(output_Win32_displaywindow,x,y,output_Win32_width,output_Win32_height,1);
		return 0;

	case WM_LBUTTONUP:
		if (output_Win32_child) break;
		if (sx == INT_MIN) break;
		sx = INT_MIN;
		return 0;

	case WM_PAINT:
		if (!output_Win32_mem) return;
		cur = 0;
		for (i = 0; i < BUFFERS; i++) {
			if (output_Win32_picnr[i] > output_Win32_picnr[cur] &&
				output_Win32_picnr[i] <= output_Win32_last &&
				output_Win32_picnr[i] <= audiopic)
					cur = i;
		}
		if (output_Win32_picnr[cur] <= audiopic) {
	                hdc = BeginPaint(hwnd, &ps);
			NOTNULL(SelectObject(output_Win32_mem, output_Win32_dcbm[cur]),
				(MSG("could not initialize display area (Select)")));
			BitBlt(hdc, 0, 0, output_Win32_width, output_Win32_height, output_Win32_mem, 0, 0, SRCCOPY);
			EndPaint(hwnd, &ps);
			for (i = 0; i < BUFFERS; i++) {
				if (output_Win32_picnr[i] <= output_Win32_picnr[cur] ||
					output_Win32_picnr[i] > output_Win32_last ||
					output_Win32_picnr[i] > audiopic+25)
						output_Win32_picnr[i] = -1;
			}
		}
                return 0;

	case WM_CHAR:
		if (output_Win32_child) break;
		if ((TCHAR)wParam != 'q') break;
	case WM_CLOSE:
                if (output_Win32_child) break;
		output_Win32_abort = 1;
		DestroyWindow(output_Win32_displaywindow);
                return 0;

        case WM_DESTROY:
		output_Win32_abort = 1;
		if (output_Win32_dc) {
			ReleaseDC(output_Win32_displaywindow, output_Win32_dc);
			output_Win32_dc = 0;
		}
                PostMessage(output_Win32_displaywindow,WM_QUIT,0,0);
		output_Win32_displaywindow = 0;
                break;
		
		
        }
	return (DefWindowProc(hwnd, message, wParam, lParam));
}

static int lastpic = -1;
void output_Win32_start(const char *unused)
{
	WNDCLASS wc;
	DWORD tid;

	output_Win32_abort = 0;
	output_Win32_last = -1;
	audiopic = -1;
	lastpic = -1;
	output_Win32_bytes = color_bytes(color_val(COLOR),output_Win32_width,output_Win32_height);
        output_Win32_hInst = GetModuleHandleA(0);

	wc.style = CS_BYTEALIGNCLIENT;
	wc.lpfnWndProc   = output_Win32_WinProc; 
	wc.cbClsExtra    = 0 ;
	wc.cbWndExtra    = 0 ;
	wc.hInstance     = output_Win32_hInst;
	wc.hIcon         = LoadIcon(NULL, IDI_EXCLAMATION);
        wc.hCursor       = LoadCursor(NULL, IDC_CROSS);
	wc.hbrBackground = NULL;
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = CLASSNAME;
	ASSERT(RegisterClass(&wc) != FALSE,(MSG("could not register class")));

        NOTNULL(output_Win32_hth = CreateThread(NULL,0,output_Win32_messageloop,output_Win32_hInst,0,&tid),
		(MSG("could not create message thread")));
	while (!output_Win32_init) Sleep(10);
}

void output_Win32_stop(void)
{
        if (output_Win32_displaywindow) {
                DestroyWindow(output_Win32_displaywindow);
        }
        if (output_Win32_hth) {
		TerminateThread(output_Win32_hth,0);
		output_Win32_hth = 0;
	}
	if (output_Win32_init) {
		int i;
		SelectObject(output_Win32_mem, output_Win32_old);
		for (i = 0; i < BUFFERS; i++) {
			DeleteObject(output_Win32_dcbm[i]);
		}
		DeleteDC(output_Win32_mem);
		output_Win32_mem = 0;
		output_Win32_init = 0;
	}
	SAFE_FREE(output_Win32_displaybitmap);
	UnregisterClass(CLASSNAME, output_Win32_hInst);
	output_Win32_audio_sync = 0;
}
 
int output_Win32_write(unsigned char *buffer, int picnr)
{
	int i, cur, lp = -1;

	if (output_Win32_abort) {
		ERROR(NOERR("user abort"));
		END(0);
	}

	cur = 0;
	for (i = 1; i < BUFFERS; i++) {
		if (output_Win32_picnr[i] < output_Win32_picnr[cur]) {
			cur = i;
			if (output_Win32_picnr[i] > 0) lp = i;
		}
	}
	if (output_Win32_audio_sync) while (output_Win32_picnr[lp] >= audiopic) Sleep(10);

	memcpy(output_Win32_bitmap[cur],buffer,output_Win32_bytes);
	output_Win32_picnr[cur] = picnr;
	output_Win32_last = picnr;
	if (output_Win32_audio_sync <= 0) output_Win32_display(picnr);

	output_Win32_disp();
	return 1;
}

void output_Win32_display(int picnr)
{
	audiopic = picnr;
	output_Win32_disp();
}

void output_Win32_disp()
{
        RECT rect;
	char msg[80];

	if (lastpic == audiopic) return;
	lastpic = audiopic;
        GetClientRect(output_Win32_displaywindow, &rect);
        InvalidateRect(output_Win32_displaywindow, &rect, FALSE);
	sprintf(msg,"displaying frame [%i]",audiopic);
	(*transport_otrace)(msg);
}

void output_Win32_setsync(int i)
{
	output_Win32_audio_sync = i;
}
#endif
