/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#ifdef HAVE_LINUX_FB_H
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <limits.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <linux/fb.h>
#ifdef HAVE_LIBGPM
#include <gpm.h>
#endif

#define TRANSPORT_INTERNAL
#include "version.h"
#include "error.h"
#include "configuration.h"
#include "color.h"
#include "transport.h"
#define SYMBOL(x) output_fb_ ## x

/************ Prototypes ************/

void output_fb_start(const char *unused);
void output_fb_stop(void);
int output_fb_write(unsigned char *buffer, int picnr);

/************************************/

int output_fb_fmts() { return 1; }
const char *output_fb_fmt(int nr) { return "fb"; }

char *output_fb_buffer = NULL;
__u16 output_fb_red[256];
__u16 output_fb_green[256];
__u16 output_fb_blue[256];
struct fb_cmap output_fb_cmap = {0, 0, output_fb_red, output_fb_green, output_fb_blue, NULL};
const char *output_fb_color_fmt = "";
struct fb_var_screeninfo output_fb_old, output_fb_low;
int output_fb_size = 0;
int output_fb_hsize = 0;
int output_fb_vsize = 0;
int output_fb_fd = -1;
int output_fb_bytes = 0;
int output_fb_line = 0;
int output_fb_width = 0, output_fb_height = 0;
int output_fb_xoff = 0, output_fb_yoff = 0;
char output_fb_device[PATH_MAX] = "/dev/fb0";

START_DATA
INTOPTION(width,0,INT_MAX);
INTOPTION(height,0,INT_MAX);
INTOPTION(xoff,0,INT_MAX);
INTOPTION(yoff,0,INT_MAX);
OPTION(color_fmt) NOSET O_GET {
	struct fb_var_screeninfo var;
	int fd = -1;

	memset(&var,0,sizeof(var));
	SAFE_OPEN(fd,output_fb_device,O_RDONLY);
	NOTNEG1(ioctl(fd,FBIOGET_VSCREENINFO,&var),(ERRMSG("FBIOGET_VSCREENINFO")));
	if (var.grayscale) {
		ASSERT(var.bits_per_pixel == 8,(MSG("grayscale only supported for 8 bpp")));
		output_fb_color_fmt = "GREY8";
	} else {
		switch (var.bits_per_pixel) {
		case 8:
			output_fb_color_fmt = "PAL8";
			break;
		case 16:
			output_fb_color_fmt = "RGB15";
			break;
		case 24:
			output_fb_color_fmt = "RGB24S";
			break;
		default:
			ASSERT(0,(MSG("%i bpp not supported"),var.bits_per_pixel));
		}
	}
	strcpy(value,output_fb_color_fmt);
}
STROPTION(device);
END_DATA

void output_fb_start(const char *unused)
{
	int i = 0, r, g, b;
	int tg[9] = {0,32,64,96,128,160,192,224,255};
	int trb[5] = {0,64,128,192,255};

#ifdef HAVE_LIBGPM
	{
		struct Gpm_Connect conn;
		conn.eventMask = GPM_MOVE|GPM_DOWN;
		conn.defaultMask = ~conn.eventMask;
		conn.minMod = 4; // control only
		conn.maxMod = 4;
		gpm_zerobased = 1;
		if (getuid()) Gpm_Open(&conn,0);
		else {
			setsid();
			Gpm_Open(&conn,-1);
		}
		fcntl(gpm_fd,F_SETFL,O_NONBLOCK);
	}
#endif

	SAFE_OPEN(output_fb_fd,output_fb_device,O_RDWR);
	NOTNEG1(ioctl(output_fb_fd,FBIOGET_VSCREENINFO,&output_fb_old),(ERRMSG("FBIOGET_VSCREENINFO")));
	output_fb_old.xoffset = 0;
	output_fb_old.yoffset = 0;
	output_fb_old.activate = FB_ACTIVATE_NOW|FB_ACTIVATE_VBL|FB_ACTIVATE_ALL|FB_CHANGE_CMAP_VBL;    
	NOTNEG1(ioctl(output_fb_fd,FBIOPAN_DISPLAY,&output_fb_old),(ERRMSG("FBIOPAN_DISPLAY")));
	memcpy(&output_fb_low,&output_fb_old,sizeof(output_fb_low));

	output_fb_low.xres = 768;
	output_fb_low.yres = 576;
	output_fb_low.xres_virtual = 768;
	output_fb_low.yres_virtual = 576;

	output_fb_low.pixclock = 16384;
	output_fb_low.left_margin = 160;
	output_fb_low.right_margin = 64;
	output_fb_low.upper_margin = 30;
	output_fb_low.lower_margin = 4;
	output_fb_low.hsync_len = 64;
	output_fb_low.vsync_len = 6;
	output_fb_low.sync = 0;
	output_fb_low.vmode = FB_VMODE_NONINTERLACED;

	output_fb_bytes = color_bytes(color_val(output_fb_color_fmt),output_fb_width,1);
	output_fb_line = color_bytes(color_val(output_fb_color_fmt),output_fb_old.xres_virtual,1);
	output_fb_size = output_fb_old.xres_virtual*output_fb_old.yres_virtual*output_fb_old.bits_per_pixel/8;
	output_fb_hsize = output_fb_old.xres_virtual;
	output_fb_vsize = output_fb_old.yres;

	NOTNEG1((int)(output_fb_buffer=mmap(0,output_fb_size,PROT_READ|PROT_WRITE,MAP_SHARED,output_fb_fd,0)),
		(ERRMSG("mmap %i bytes"),output_fb_size));
	
	if (output_fb_old.bits_per_pixel == 8) {
		i = 0;
		output_fb_cmap.start = 16;
		output_fb_cmap.len = 250;
		for (g=0; g<9; g++) for (r=0; r<5; r++) for (b=0;b<5;b++) {
			output_fb_red[i] = trb[r]*256;
			output_fb_green[i] = tg[g]*256;
			output_fb_blue[i] = trb[b]*256;
			i++;
		}
	} else if (output_fb_old.bits_per_pixel == 16) {
		output_fb_cmap.start = 0;
		output_fb_cmap.len = 32;
		for (i = 1; i <= 32; i++) {
			output_fb_red[i-1] = i*2048-1;
			output_fb_green[i-1] = i*2048-1;
			output_fb_blue[i-1] = i*2048-1;
		}
	} else if (output_fb_old.bits_per_pixel == 24) {
		output_fb_cmap.start = 0;
		output_fb_cmap.len = 256;
		for (i = 1; i <= 256; i++) {
			output_fb_red[i-1] = 256*i-1;
			output_fb_green[i-1] = 256*i-1;
			output_fb_blue[i-1] = 256*i-1;
		}
	}
	ioctl(output_fb_fd,FBIOPUTCMAP,&output_fb_cmap);

}

void output_fb_stop(void)
{
	if (output_fb_buffer && output_fb_buffer != ((void *)-1)) {
		munmap(output_fb_buffer,output_fb_size);
		output_fb_buffer = NULL;
	}
	if (output_fb_fd >= 0) {
		NOTNEG1(ioctl(output_fb_fd,FBIOPUT_VSCREENINFO,&output_fb_old),(ERRMSG("FBIOSET_VSCREENINFO")));
		write(0,"\e[?25h\e[?8c",11);
		close(output_fb_fd);
		output_fb_fd = -1;
	}

#ifdef HAVE_LIBGPM
	Gpm_Close();
#endif
}

int output_fb_write(unsigned char *buffer, int picnr)
{
	char *ptr;
	char msg[PATH_MAX];
	int i = output_fb_height;
#ifdef HAVE_LIBGPM
	Gpm_Event ev;
	fd_set set;
	struct timeval nul = {0,0};
	FD_ZERO(&set);
	FD_SET(gpm_fd,&set);

	if (select(gpm_fd+1,&set,NULL,NULL,&nul) > 0 && Gpm_GetEvent(&ev) > 0) {
		while (Gpm_GetEvent(&ev) > 0);
		if (ev.buttons & GPM_B_LEFT) {
			ioctl(output_fb_fd,FBIOPUTCMAP,&output_fb_cmap);
		} else if (ev.buttons & GPM_B_MIDDLE) {
			NOTNEG1(ioctl(output_fb_fd,FBIOPUT_VSCREENINFO,&output_fb_low),(ERRMSG("FBIOSET_VSCREENINFO")));
			write(0,"\e[?25l\e[?1c",11);
			output_fb_line = color_bytes(color_val(output_fb_color_fmt),output_fb_low.xres_virtual,1);
			output_fb_size = output_fb_low.xres_virtual*output_fb_low.yres_virtual*output_fb_low.bits_per_pixel/8;
			output_fb_hsize = output_fb_low.xres_virtual;
			output_fb_vsize = output_fb_low.yres;
			output_fb_xoff = 0;
			output_fb_yoff = 0;
		} else if (ev.buttons & GPM_B_RIGHT) {
			NOTNEG1(ioctl(output_fb_fd,FBIOPUT_VSCREENINFO,&output_fb_old),(ERRMSG("FBIOSET_VSCREENINFO")));
			write(0,"\e[?25h\e[?8c",11);
			output_fb_line = color_bytes(color_val(output_fb_color_fmt),output_fb_old.xres_virtual,1);
			output_fb_size = output_fb_old.xres_virtual*output_fb_old.yres_virtual*output_fb_old.bits_per_pixel/8;
			output_fb_hsize = output_fb_old.xres_virtual;
			output_fb_vsize = output_fb_old.yres;
			output_fb_xoff = 0;
			output_fb_yoff = 0;
		} else {
			int x = (output_fb_hsize*ev.x)/gpm_mx;
			int y = (output_fb_vsize*ev.y)/gpm_my;
			output_fb_xoff = x-(output_fb_width/2);
			output_fb_yoff = y-(output_fb_height/2);
			if (output_fb_xoff+output_fb_width >= output_fb_hsize) output_fb_xoff = output_fb_hsize-output_fb_width;
			if (output_fb_yoff+output_fb_height >= output_fb_vsize) output_fb_yoff=output_fb_vsize-output_fb_height;
			if (output_fb_xoff < 0) output_fb_xoff = 0;
			if (output_fb_yoff < 0) output_fb_yoff = 0;
		}
	}
#endif

	ptr = output_fb_buffer + (output_fb_yoff*output_fb_line) + (output_fb_xoff*(output_fb_line/output_fb_hsize));
	while (i-- > 0) {
		memcpy(ptr,buffer,output_fb_bytes);
		ptr += output_fb_line;
		buffer += output_fb_bytes;
	}
	sprintf(msg,"displaying frame [%i]",picnr);
	(*transport_otrace)(msg);
	return 1;
}


#endif
