/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "error.h"
#include "configuration.h"
#include "read_sound.h"
#include "transport.h"
#define SYMBOL(x) read_sound_wav_ ## x

/************ Prototypes ************/

void read_sound_wav_start(const char *name);
void read_sound_wav_setnext(int fbnr, int picnr);
int read_sound_wav_get(int *fbnr, int *picnr);
void read_sound_wav_stop();

/************************************/

/* byte order stuff */
#ifndef WORDS_BIGENDIAN
# define cpu_to_le32(x) (x)
# define cpu_to_le16(x) (x)
# define le32_to_cpu(x) (x)
# define le16_to_cpu(x) (x)
#else
#error "byte order currently not supported, perhaps you want to fix it right here?"
#endif
/* other byte orders still missing */



/* ---------------------------------------------------------------------- */
/* *.wav I/O stolen from krecord, which in turn stole from cdda2wav */
/* Copyright (C) by Heiko Eissfeldt */

typedef unsigned char  BYTE;
typedef unsigned short WORD;
typedef unsigned long  DWORD;
typedef unsigned long  FOURCC;	/* a four character code */

/* flags for 'wFormatTag' field of WAVEFORMAT */
#define WAVE_FORMAT_PCM 1

/* MMIO macros */
#define mmioFOURCC(ch0, ch1, ch2, ch3) \
  ((DWORD)(BYTE)(ch0) | ((DWORD)(BYTE)(ch1) << 8) | \
  ((DWORD)(BYTE)(ch2) << 16) | ((DWORD)(BYTE)(ch3) << 24))

#define FOURCC_RIFF	mmioFOURCC ('R', 'I', 'F', 'F')
#define FOURCC_WAVE	mmioFOURCC ('W', 'A', 'V', 'E')
#define FOURCC_FMT	mmioFOURCC ('f', 'm', 't', ' ')
#define FOURCC_DATA	mmioFOURCC ('d', 'a', 't', 'a')

struct CHUNKHDR {
    FOURCC ckid;		/* chunk ID */
    DWORD dwSize; 	        /* chunk size */
};

/* simplified Header for standard WAV files */
struct WAVEHDR {
    struct CHUNKHDR chkRiff;
    FOURCC fccWave;
    struct CHUNKHDR chkFmt;
    WORD wFormatTag;	   /* format type */
    WORD nChannels;	   /* number of channels (i.e. mono, stereo, etc.) */
    DWORD nSamplesPerSec;  /* sample rate */
    DWORD nAvgBytesPerSec; /* for buffer estimation */
    WORD nBlockAlign;	   /* block size of data */
    WORD wBitsPerSample;
    struct CHUNKHDR chkData;
} read_sound_wav_fileheader;

#define FMT_8BIT       1          /* unsigned */
#define FMT_16BIT      2          /* signed - native byte order */

/* ---------------------------------------------------------------------- */

int read_sound_wav_bufsize;

int read_sound_wav_rate     = 44100;
int read_sound_wav_bits     = 16;
int read_sound_wav_channels = 2;

unsigned long read_sound_wav_align;

START_DATA
INTINFO(rate);
INTINFO(bits);
INTINFO(channels);
INTINFO(bufsize);
END_DATA

void read_sound_wav_start(const char *name)
{
	transport_srstart(name);
	transport_sread(&read_sound_wav_fileheader,sizeof(struct WAVEHDR));

	ASSERT(read_sound_wav_fileheader.chkRiff.ckid == cpu_to_le32(FOURCC_RIFF) &&
	      read_sound_wav_fileheader.fccWave         == cpu_to_le32(FOURCC_WAVE) &&
	      read_sound_wav_fileheader.chkFmt.ckid     == cpu_to_le32(FOURCC_FMT) &&
	      read_sound_wav_fileheader.chkFmt.dwSize   == cpu_to_le32(16) &&
	      read_sound_wav_fileheader.wFormatTag      == cpu_to_le16(WAVE_FORMAT_PCM) &&
	      read_sound_wav_fileheader.chkData.ckid    == cpu_to_le32(FOURCC_DATA),
	      (MSG("Sorry, only very simple WAV files (like those created by bttvgrab) supported")));

	read_sound_wav_channels = le16_to_cpu(read_sound_wav_fileheader.nChannels);
	read_sound_wav_rate = le32_to_cpu(read_sound_wav_fileheader.nSamplesPerSec);
	read_sound_wav_bits = le16_to_cpu(read_sound_wav_fileheader.wBitsPerSample);
	
	read_sound_wav_bufsize = read_sound_wav_bits/8 * read_sound_wav_rate/25 * read_sound_wav_channels;
}

void read_sound_wav_setnext(int fbnr, int picnr)
{
}

int read_sound_wav_get(int *fbnr, int *picnr)
{
	static int nextpic = 0;
	*fbnr = 0;
	transport_srframe(picnr);
	transport_srlseek((read_sound_wav_bufsize*(*picnr-nextpic)),SEEK_CUR);
	transport_sread(read_sound_buf[0],read_sound_wav_bufsize);
	nextpic = *picnr+1;
	return read_sound_wav_bufsize;
}

void read_sound_wav_stop()
{
	transport_srstop();
}
