/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <errno.h>
#include "transport.h"
#include "sound.h"
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) sound_wav_ ## x
#ifdef __CYGWIN__
#define PTH_T(x) &x
#else
#define PTH_T(x) x
#endif

/************ Prototypes ************/

void sound_wav_start(const char *name);
void sound_wav_write(unsigned char *buffer, int picnr, int len);
void sound_wav_stop();

/************************************/

/* byte order stuff */
#ifndef WORDS_BIGENDIAN
# define cpu_to_le32(x) (x)
# define cpu_to_le16(x) (x)
# define le32_to_cpu(x) (x)
# define le16_to_cpu(x) (x)
#else
#error "byte order currently not supported, perhaps you want to fix it right here?"
#endif
/* other byte orders still missing */



/* ---------------------------------------------------------------------- */
/* *.wav I/O stolen from krecord, which in turn stole from cdda2wav */
/* Copyright (C) by Heiko Eissfeldt */

typedef unsigned char  BYTE;
typedef unsigned short WORD;
typedef unsigned long  DWORD;
typedef unsigned long  FOURCC;	/* a four character code */

/* flags for 'wFormatTag' field of WAVEFORMAT */
#define WAVE_FORMAT_PCM 1

/* MMIO macros */
#define mmioFOURCC(ch0, ch1, ch2, ch3) \
  ((DWORD)(BYTE)(ch0) | ((DWORD)(BYTE)(ch1) << 8) | \
  ((DWORD)(BYTE)(ch2) << 16) | ((DWORD)(BYTE)(ch3) << 24))

#define FOURCC_RIFF	mmioFOURCC ('R', 'I', 'F', 'F')
#define FOURCC_WAVE	mmioFOURCC ('W', 'A', 'V', 'E')
#define FOURCC_FMT	mmioFOURCC ('f', 'm', 't', ' ')
#define FOURCC_DATA	mmioFOURCC ('d', 'a', 't', 'a')

struct CHUNKHDR {
    FOURCC ckid;		/* chunk ID */
    DWORD dwSize; 	        /* chunk size */
};

/* simplified Header for standard WAV files */
struct WAVEHDR {
    struct CHUNKHDR chkRiff;
    FOURCC fccWave;
    struct CHUNKHDR chkFmt;
    WORD wFormatTag;	   /* format type */
    WORD nChannels;	   /* number of channels (i.e. mono, stereo, etc.) */
    DWORD nSamplesPerSec;  /* sample rate */
    DWORD nAvgBytesPerSec; /* for buffer estimation */
    WORD nBlockAlign;	   /* block size of data */
    WORD wBitsPerSample;
    struct CHUNKHDR chkData;
} sound_wav_fileheader;

#define FMT_8BIT       1          /* unsigned */
#define FMT_16BIT      2          /* signed - native byte order */

/* ---------------------------------------------------------------------- */

int sound_wav_rate     = 44100;
int sound_wav_bits     = 16;
int sound_wav_channels = 2;

unsigned long sound_wav_align;
int sound_wav_size = 0;
int sound_wav_init_p = 0;

START_DATA
INTOPTION(rate,4000,48000);
INTOPTION(bits,8,16);
INTOPTION(channels,1,4);
END_DATA

void sound_wav_start(const char *name)
{
	sound_wav_align = sound_wav_channels * ((sound_wav_bits + 7) / 8);
	
	sound_wav_fileheader.chkRiff.ckid    = cpu_to_le32(FOURCC_RIFF);
	sound_wav_fileheader.chkRiff.dwSize  = cpu_to_le32(sizeof(struct WAVEHDR) - sizeof(struct CHUNKHDR));
	sound_wav_fileheader.fccWave         = cpu_to_le32(FOURCC_WAVE);
	sound_wav_fileheader.chkFmt.ckid     = cpu_to_le32(FOURCC_FMT);
	sound_wav_fileheader.chkFmt.dwSize   = cpu_to_le32(16);
	sound_wav_fileheader.wFormatTag      = cpu_to_le16(WAVE_FORMAT_PCM);
	sound_wav_fileheader.chkData.ckid    = cpu_to_le32(FOURCC_DATA);
	sound_wav_fileheader.chkData.dwSize  = cpu_to_le32(0);
	sound_wav_fileheader.nChannels       = cpu_to_le16(sound_wav_channels);
	sound_wav_fileheader.nSamplesPerSec  = cpu_to_le32(sound_wav_rate);
	sound_wav_fileheader.nAvgBytesPerSec = cpu_to_le32(sound_wav_align * sound_wav_rate);
	sound_wav_fileheader.nBlockAlign     = cpu_to_le16(sound_wav_align);
	sound_wav_fileheader.wBitsPerSample  = cpu_to_le16(sound_wav_bits);
	sound_wav_size = 0;

	transport_swstart(name);
	transport_swrite(&sound_wav_fileheader,sizeof(struct WAVEHDR));
}

#define NORMSIZE ((sound_wav_rate/25) * sound_wav_channels * (sound_wav_bits/8))
#define TRIGGERSIZE (5*NORMSIZE)
#define ADAPTSIZE (NORMSIZE/2)

void sound_wav_write(unsigned char *buffer, int picnr, int len)
{
	transport_swframe(picnr);
	if (len > NORMSIZE) len = NORMSIZE;

	transport_swrite(buffer,len);
	if (len < NORMSIZE) transport_swlseek(NORMSIZE-len,SEEK_CUR);
	sound_wav_size += NORMSIZE;
}

void sound_wav_stop()
{
	sound_wav_fileheader.chkRiff.dwSize = cpu_to_le32(sound_wav_size + sizeof(struct WAVEHDR) - sizeof(struct CHUNKHDR));
	sound_wav_fileheader.chkData.dwSize = cpu_to_le32(sound_wav_size);
	transport_swlseek(0,SEEK_SET);
	transport_swrite(&sound_wav_fileheader,sizeof(struct WAVEHDR));

	transport_swstop();
}
