/************************************************************************
 *
 *  coder.c, main coding engine of tmn (TMN encoder)
 *
 *  Copyright (C) 1997  University of BC, Canada
 *
 *  Contacts:
 *  Michael Gallant                   <mikeg@ee.ubc.ca>
 *  Guy Cote                          <guyc@ee.ubc.ca>
 *  Berna Erol                        <bernae@ee.ubc.ca>
 *
 *  UBC Image Processing Laboratory   http://www.ee.ubc.ca/image
 *  2356 Main Mall                    tel.: +1 604 822 4051
 *  Vancouver BC Canada V6T1Z4        fax.: +1 604 822 5949
 *
 ************************************************************************/

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any
 * license fee or royalty on an "as is" basis. The University of
 * British Columbia disclaims any and all warranties, whether
 * express, implied, or statuary, including any implied warranties
 * or merchantability or of fitness for a particular purpose.  In no
 * event shall the copyright-holder be liable for any incidental,
 * punitive, or consequential damages of any kind whatsoever arising
 * from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs
 * and user's customers, employees, agents, transferees, successors,
 * and assigns.
 *
 * The University of British Columbia does not represent or warran
 * that the programs furnished hereunder are free of infringement of
 * any third-party patents.
 *
 * Commercial implementations of H.263, including shareware, are
 * subject to royalty fees to patent holders.  Many of these patents
 * are general enough such that they are unavoidable regardless of
 * implementation design.
 *
*/

#include "sim.h"
#include "../error.h"

/**********************************************************************
 *
 *	Name:         CodeOneEI
 *	Description:  codes one image intra
 *	
 *	Input:        pointer to image, QP
 *
 *	Returns:      pointer to reconstructed image
 *	Side effects: memory is allocated to recon image
 *
 *	Date: 940110  Author: Michael Gallant	mikeg@ee.ubc.ca
 *
 ***********************************************************************/


void output_H263_CodeOneEI(PictImage *curr_image, PictImage *pr,
			   int QP, Bits *bits, Pict *pic, PictImage *recon) {
	MB_Structure *diff = (MB_Structure *)malloc(sizeof(MB_Structure));
	MB_Structure *recon_data_ei = (MB_Structure *)malloc(sizeof(MB_Structure));
	PictImage *base_recon=NULL, *base_recon_edge=NULL;
	int *qcoeff, *pcoeff=NULL, *rcoeff, *coeff;
	int Mode = MODE_INTER;
	int CBP;
	int i,j,k;
	int *store_coeff=NULL, *store_rcoeff=NULL; /* used to do prediction in advanced intra coding */
	int newgob = 0;
	int ei_prediction_type;
	unsigned char *base_ipol, *bi;
	int advanced_temporarily_off = DEF_AIC_MODE;
	
	/* buffer control vars */
	float QP_cumulative = (float)0.0;
	int abs_mb_num = 0;
	int QP_new, QP_prev, QP_xmitted=QP;
	
	output_H263_ZeroBits(bits);

	output_H263_GenerateFrameAndInterpolatedImages(pr, pic, &base_ipol, &base_recon, &bi, &base_recon_edge);

	if (output_H263_adv_pred) {
		/* Turn off advanced coding since there can be only 1
		 * motion vector for B frames of IPB. */
		advanced_temporarily_off = YES;
		output_H263_overlapping_MC = OFF;
		output_H263_adv_pred = OFF;
		output_H263_use_4mv = OFF;
	}

	QP_new = QP_xmitted = QP_prev = QP; /* Copy the passed value of QP */

	for ( j = 0; j < output_H263_lines/MB_SIZE; j++) {
		newgob = 0;

		if (j == 0) {
			if (output_H263_advanced_intra_coding) {
				/* store the coeff for the frame */
				SAFE_CALLOC(store_coeff,384*(output_H263_pels/MB_SIZE)*(output_H263_lines/MB_SIZE), sizeof(int));
				SAFE_CALLOC(store_rcoeff,384*(output_H263_pels/MB_SIZE)*(output_H263_lines/MB_SIZE), sizeof(int));
			}

			pic->QUANT = QP_new;

			bits->header += output_H263_CountBitsPicture(pic);

			QP_xmitted = QP_prev = QP_new;
		} else if (pic->use_gobsync && j%pic->use_gobsync == 0) {
			bits->header += output_H263_CountBitsGOB(j,QP_new,pic); /* insert gob sync */
			QP_xmitted = QP_prev = QP_new;
			newgob = 1;
		}

		for ( i = 0; i < output_H263_pels/MB_SIZE; i++) {
			pic->MB = i + j * (output_H263_pels/MB_SIZE);

			/* store the QP for every macroblock */
			output_H263_quant_map[j+1][i+1] = QP_xmitted;

			SAFE_MALLOC(rcoeff ,sizeof(int)*384);

			/* Predict true B-MB */
			diff = output_H263_Predict_EI( curr_image, base_recon, base_ipol, i*MB_SIZE, j*MB_SIZE, &ei_prediction_type, &Mode, pic->RTYPE);

			if (EI_INTRA_PREDICTION == ei_prediction_type) {
				output_H263_FillLumBlock(i*MB_SIZE, j*MB_SIZE, curr_image, diff);
				output_H263_FillChromBlock(i*MB_SIZE, j*MB_SIZE, curr_image, diff);
			}

			SAFE_MALLOC(qcoeff,sizeof(int)*384);
			coeff = output_H263_MB_Encode(diff);

			if (output_H263_advanced_intra_coding) {	
				if (!(Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) {
					for (k=0;k<6;k++) {
						/* store default coeff if non-intra macroblock */
						store_coeff[(i + j * output_H263_pels/MB_SIZE) * 384 + k * 64] = 1024;
						store_rcoeff[(i + j * output_H263_pels/MB_SIZE) * 384 + k * 64] = 1024;
					}
					for (k=0;k<6;k++) output_H263_Quant_blk(coeff,qcoeff,QP_xmitted, Mode,k);
					CBP = output_H263_FindCBP(qcoeff, Mode, 64);
					if (CBP == 0 && (Mode == MODE_INTER || Mode == MODE_INTER_Q)) {
						output_H263_ZeroMBlock(diff);
					} else {
						for (k=0;k<6;k++) {
							output_H263_Quant_blk(coeff,qcoeff,QP_xmitted, Mode,k);
							output_H263_Dequant(qcoeff, rcoeff, QP_xmitted, Mode,k);
						}
						output_H263_MB_Decode(rcoeff, diff);
					}
				} else {
					SAFE_MALLOC(pcoeff,sizeof(int)*384);

					/* store the quantized DCT coefficients */
					memcpy( (void *) (store_coeff + (i + j*output_H263_pels/MB_SIZE)*384), (void *) coeff, sizeof(int) * 384);

					/* Do Intra mode prediction */
					pic->Intra_Mode = output_H263_Choose_Intra_Mode(pcoeff, store_coeff, i, j, newgob);

					for (k=0;k<6;k++) {
						output_H263_Intra_AC_DC_Encode(coeff, store_rcoeff, pic->Intra_Mode, i, j, newgob,k);
						output_H263_Quant_blk(coeff,pcoeff,QP_xmitted,Mode,k);
						output_H263_Dequant(pcoeff, rcoeff, QP_xmitted, Mode,k);
						output_H263_Intra_AC_DC_Decode(rcoeff, store_rcoeff, pic->Intra_Mode, i, j, newgob,k);
					}
					output_H263_MB_Decode(rcoeff, diff);
					CBP = output_H263_FindCBP(pcoeff,Mode,64);
				}
			} else {
				for (k=0;k<6;k++) output_H263_Quant_blk(coeff,qcoeff,QP_xmitted, Mode,k);
				CBP = output_H263_FindCBP(qcoeff, Mode, 64);
				if (CBP == 0 && (Mode == MODE_INTER || Mode == MODE_INTER_Q)) {
					output_H263_ZeroMBlock(diff);
				} else {
					for (k=0;k<6;k++) output_H263_Dequant(qcoeff, rcoeff, QP_xmitted, Mode,k);
					output_H263_MB_Decode(rcoeff, diff);
				}
			}

			recon_data_ei = output_H263_MB_Recon_EI( base_recon, base_ipol, diff, i*MB_SIZE,j*MB_SIZE, ei_prediction_type, pic->RTYPE);

			output_H263_Clip(recon_data_ei);
			free(diff);
			free(coeff);

			if (!CBP) output_H263_intra_refresh[j+1][i+1] += 1;

			if ((CBP==0) && (EI_UPWARD_PREDICTION == ei_prediction_type) && (Mode == MODE_INTER) && (0 == pic->DQUANT) ) {
				output_H263_coded_map[j+1][i+1] = 0;
				output_H263_quant_map[j+1][i+1] = 0;

				output_H263_CountBitsScalMB(Mode, 1, CBP, 0, pic, bits, ei_prediction_type, 0);
			} else {
				output_H263_CountBitsScalMB(Mode,0,CBP,0,pic,bits,ei_prediction_type, 0);

				if (MODE_INTER == Mode || MODE_INTER_Q == Mode) {
					output_H263_coded_map[j+1][i+1] = 1;
					output_H263_quant_map[j+1][i+1] = QP_xmitted;
					bits->no_inter++;
				} else {
					/* MODE_INTRA or MODE_INTRA_Q */
					output_H263_coded_map[j+1][i+1] = 2;
					output_H263_quant_map[j+1][i+1] = QP_xmitted;
					output_H263_intra_refresh[j+1][i+1] = 0;
					bits->no_intra++;
				}

				if ( (Mode == MODE_INTRA || Mode == MODE_INTRA_Q) && output_H263_advanced_intra_coding ) {
					output_H263_Scan(pcoeff,pic->Intra_Mode);
					output_H263_CountBitsCoeff(pcoeff, Mode, CBP, bits, 64);
				} else if (CBP || Mode == MODE_INTRA || Mode == MODE_INTRA_Q) {
					output_H263_Scan(qcoeff,0);
					output_H263_CountBitsCoeff(qcoeff, Mode, CBP, bits, 64);
				}
			}
			QP_prev = QP_xmitted;

			abs_mb_num++;
			QP_cumulative += QP_xmitted;
			output_H263_ReconImage(i,j,recon_data_ei,recon);

			free(qcoeff);
			free(recon_data_ei);

			if (output_H263_advanced_intra_coding && (Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) free(pcoeff);
		}
	}

	pic->QP_mean = QP_cumulative/(float)abs_mb_num;

  /* Free memory */
	free(base_recon);
	output_H263_FreeImage(base_recon_edge);
	free(bi);

	if (output_H263_advanced_intra_coding) {
		free(store_coeff);
		free(store_rcoeff);
	}

	if (advanced_temporarily_off) {
		output_H263_overlapping_MC = ON;
		output_H263_adv_pred = ON;
		output_H263_use_4mv = ON;
		advanced_temporarily_off = NO;
	}

	return;
}

/***********************************************************************
 *
 *	Name:           Predict_EI
 *	Description:    Predicts macroblock in EI picture
 *
 *	Input:	        pointers to current frame, base recon. base
 *                      interpolated, pos. in image, MV data,
 *                      and MB prediction type.
 *	Returns:        pointer to differential MB data after prediction
 *	Side effects:
 *
 *	Date: 970831    Author: Michael Gallant ---mikeg@ee.ubc.ca
 *
 ***********************************************************************/
MB_Structure *output_H263_Predict_EI( PictImage *curr_image, PictImage *base_recon,
				      unsigned char *base_ipol, int x, int y,
				      int *prediction_type , int *mode, int RTYPE) {
	int curr[16][16];
	MB_Structure *upward_pred = (MB_Structure *)malloc(sizeof(MB_Structure));
	MB_Structure *pred_error = (MB_Structure *) malloc (sizeof (MB_Structure));
	MotionVector ZERO = {0,0,0,0,0};
	int i,j;
	int SADup;

	/* Find MB in current image */
	output_H263_FindMB (x, y, curr_image->lum, curr);

	/* Find predicted MB for current MB (upward prediction so MV=0 */
	output_H263_FindPred (x, y, &ZERO, base_ipol, &upward_pred->lum[0][0], 16, 0);

	SADup = output_H263_SAD_MB_integer (&curr[0][0], &upward_pred->lum[0][0], 16, INT_MAX);
	SADup -= 150;

	*(mode) = output_H263_ChooseMode( curr_image->lum, x, y, SADup);

	if (MODE_INTRA == *(mode) || MODE_INTRA_Q == *(mode)) {
		*(prediction_type) = EI_INTRA_PREDICTION;
	} else {
		*(prediction_type) = EI_UPWARD_PREDICTION;
	}

	switch (*(prediction_type)) {
	case EI_UPWARD_PREDICTION:

		output_H263_DoPredChrom_P (x, y, 0, 0, curr_image, base_recon, upward_pred, pred_error, RTYPE);

		for (j = 0; j < MB_SIZE; j++) {
			for (i = 0; i < MB_SIZE; i++) {
				pred_error->lum[j][i] = *(curr_image->lum + x + i + (y + j) * output_H263_pels) - upward_pred->lum[j][i];
			}
		}

		break;

	case EI_INTRA_PREDICTION:

		break;

	default:

		break;

	}

	free (upward_pred);

	return pred_error;
}

/***********************************************************************
 *
 *	Name:           MB_Recon_EI
 *	Description:    Reconstructs EI macroblocks
 *
 *	Input:	
 *	Returns:        pointer to differential MB data after prediction
 *	Side effects:
 *
 *	Date: 970831    Author: Michael Gallant ---mikeg@ee.ubc.ca
 *
 ***********************************************************************/
MB_Structure *output_H263_MB_Recon_EI( PictImage *base_recon, unsigned char *base_ipol,
				       MB_Structure *diff, int x, int y,
				       int prediction_type, int RTYPE) {
	MB_Structure *upward = (MB_Structure *) malloc (sizeof (MB_Structure));
	MB_Structure *recon_data = (MB_Structure *) malloc (sizeof (MB_Structure));
	MotionVector ZERO = {0,0,0,0,0};

	switch (prediction_type) {

	case EI_UPWARD_PREDICTION:

		/* Inter 16x16 */
		output_H263_ReconLumBlock_P (x, y, &ZERO, base_ipol, &diff->lum[0][0], 16, 0);
		output_H263_ReconChromBlock_P (x, y, 0, 0, base_recon, diff, RTYPE);

		break;

	case EI_INTRA_PREDICTION:

		break;

	default:

		ERROR(INTMSG("Illegal scalable prediction type in MB_Recon_EP"));
		END(1);

		break;

	}

	memcpy(recon_data, diff, sizeof (MB_Structure));

	free(upward);

	return recon_data;
}

