/************************************************************************
 *
 *  getblk.c, DCT block decoding for tmndecode (H.263 decoder)
 *  Copyright (C) 1995, 1996  Telenor R&D, Norway
 *
 *  Contacts:
 *  Robert Danielsen                  <Robert.Danielsen@nta.no>
 *
 *  Telenor Research and Development  http://www.nta.no/brukere/DVC/
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *
 *  Copyright (C) 1997  University of BC, Canada
 *  Modified by: Michael Gallant <mikeg@ee.ubc.ca>
 *               Guy Cote <guyc@ee.ubc.ca>
 *               Berna Erol <bernae@ee.ubc.ca>
 *
 *  Contacts:
 *  Michael Gallant                   <mikeg@ee.ubc.ca>
 *
 *  UBC Image Processing Laboratory   http://www.ee.ubc.ca/image
 *  2356 Main Mall                    tel.: +1 604 822 4051
 *  Vancouver BC Canada V6T1Z4        fax.: +1 604 822 5949
 *
 ************************************************************************/

/* Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee
 * or royalty on an "as is" basis. The University of British Columbia
 * disclaims any and all warranties, whether express, implied, or
 * statuary, including any implied warranties or merchantability or of
 * fitness for a particular purpose.  In no event shall the
 * copyright-holder be liable for any incidental, punitive, or
 * consequential damages of any kind whatsoever arising from the use of
 * these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and
 * user's customers, employees, agents, transferees, successors, and
 * assigns.
 *
 * The University of British Columbia does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any
 * third-party patents.
 *
 * Commercial implementations of H.263, including shareware, are subject to
 * royalty fees to patent holders.  Many of these patents are general
 * enough such that they are unavoidable regardless of implementation
 * design.
 *
 */


/* based on mpeg2decode, (C) 1994, MPEG Software Simulation Group and
 * mpeg2play, (C) 1994 Stefan Eckart <stefan@lis.e-technik.tu-muenchen.de>
 *
 */


#include <stdio.h>
#include <stdlib.h>

#include "tmndec.h"
#include "global.h"

#define INDICES
#include "indices.h"

#define SACTABLES
#include "sactbls.h"

#include "../error.h"

/************ Prototypes ************/

void read_H263_getblock(int comp, int mode, int INTRA_AC_DC, int Mode);
void read_H263_get_sac_block (int comp, int ptype, int INTRA_AC_DC, int Mode);
RunCoef read_H263_vlc_word_decode(int symbol_word, int *last);
RunCoef read_H263_Decode_Escape_Char (int intra, int *last);
int read_H263_DecodeTCoef (int position, int intra);

/************************************/


static VLCtabI DCT3Dtab0[] = { {4225,7}, {4209,7}, {4193,7}, {4177,7}, {193,7}, {177,7}, {161,7}, {4,7}, {4161,6}, {4161,6}, {4145,6}, {4145,6}, {4129,6}, {4129,6}, {4113,6}, {4113,6}, {145,6}, {145,6}, {129,6}, {129,6}, {113,6}, {113,6}, {97,6}, {97,6}, {18,6}, {18,6}, {3,6}, {3,6}, {81,5}, {81,5}, {81,5}, {81,5}, {65,5}, {65,5}, {65,5}, {65,5}, {49,5}, {49,5}, {49,5}, {49,5}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {17,3}, {33,4}, {33,4}, {33,4}, {33,4}, {33,4}, {33,4}, {33,4}, {33,4}, {2,4}, {2,4},{2,4},{2,4}, {2,4}, {2,4},{2,4},{2,4}, };
static VLCtabI DCT3Dtab1[] = { {9,10}, {8,10}, {4481,9}, {4481,9}, {4465,9}, {4465,9}, {4449,9}, {4449,9}, {4433,9}, {4433,9}, {4417,9}, {4417,9}, {4401,9}, {4401,9}, {4385,9}, {4385,9}, {4369,9}, {4369,9}, {4098,9}, {4098,9}, {353,9}, {353,9}, {337,9}, {337,9}, {321,9}, {321,9}, {305,9}, {305,9}, {289,9}, {289,9}, {273,9}, {273,9}, {257,9}, {257,9}, {241,9}, {241,9}, {66,9}, {66,9}, {50,9}, {50,9}, {7,9}, {7,9}, {6,9}, {6,9}, {4353,8}, {4353,8}, {4353,8}, {4353,8}, {4337,8}, {4337,8}, {4337,8}, {4337,8}, {4321,8}, {4321,8}, {4321,8}, {4321,8}, {4305,8}, {4305,8}, {4305,8}, {4305,8}, {4289,8}, {4289,8}, {4289,8}, {4289,8}, {4273,8}, {4273,8}, {4273,8}, {4273,8}, {4257,8}, {4257,8}, {4257,8}, {4257,8}, {4241,8}, {4241,8}, {4241,8}, {4241,8}, {225,8}, {225,8}, {225,8}, {225,8}, {209,8}, {209,8}, {209,8}, {209,8}, {34,8}, {34,8}, {34,8}, {34,8}, {19,8}, {19,8}, {19,8}, {19,8}, {5,8}, {5,8}, {5,8}, {5,8}, };
static VLCtabI DCT3Dtab2[] = { {4114,11}, {4114,11}, {4099,11}, {4099,11}, {11,11}, {11,11}, {10,11}, {10,11}, {4545,10}, {4545,10}, {4545,10}, {4545,10}, {4529,10}, {4529,10}, {4529,10}, {4529,10}, {4513,10}, {4513,10}, {4513,10}, {4513,10}, {4497,10}, {4497,10}, {4497,10}, {4497,10}, {146,10}, {146,10}, {146,10}, {146,10}, {130,10}, {130,10}, {130,10}, {130,10}, {114,10}, {114,10}, {114,10}, {114,10}, {98,10}, {98,10}, {98,10}, {98,10}, {82,10}, {82,10}, {82,10}, {82,10}, {51,10}, {51,10}, {51,10}, {51,10}, {35,10}, {35,10}, {35,10}, {35,10}, {20,10}, {20,10}, {20,10}, {20,10}, {12,11}, {12,11}, {21,11}, {21,11}, {369,11}, {369,11}, {385,11}, {385,11}, {4561,11}, {4561,11}, {4577,11}, {4577,11}, {4593,11}, {4593,11}, {4609,11}, {4609,11}, {22,12}, {36,12}, {67,12}, {83,12}, {99,12}, {162,12}, {401,12}, {417,12}, {4625,12}, {4641,12}, {4657,12}, {4673,12}, {4689,12}, {4705,12}, {4721,12}, {4737,12}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, };
static VLCtabI INTRA_DCT3Dtab0[] = { {4099,7}, {4353,7}, {4481,7}, {4417,7}, {67,7}, {130,7}, {9,7}, {321,7}, {4098,6}, {4098,6}, {4289,6}, {4289,6}, {4225,6}, {4225,6}, {4161,6}, {4161,6}, {6,6}, {6,6}, {7,6}, {7,6}, {8,6}, {8,6}, {257,6}, {257,6}, {66,6}, {66,6}, {193,6}, {193,6}, {129,5}, {129,5}, {129,5}, {129,5}, {4,5}, {4,5}, {4,5}, {4,5}, {5,5}, {5,5}, {5,5}, {5,5}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {4097,4}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {1,2}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {2,3}, {3,4}, {3,4}, {3,4}, {3,4}, {3,4}, {3,4}, {3,4}, {3,4}, {65,4}, {65,4},{65,4},{65,4}, {65,4}, {65,4},{65,4},{65,4}, };
static VLCtabI INTRA_DCT3Dtab1[] = { {705,10}, {641,10},{4101,9}, {4101,9}, {4102,9}, {4102,9}, {4162,9}, {4162,9}, {4226,9}, {4226,9},{5057,9}, {5057,9}, {5121,9}, {5121,9}, {5249,9}, {5249,9},{5185,9}, {5185,9}, {4993,9}, {4993,9}, {13,9}, {13,9},{14,9}, {14,9}, {15,9}, {15,9}, {16,9}, {16,9},{17,9}, {17,9}, {18,9}, {18,9}, {11,9}, {11,9},{12,9}, {12,9}, {322,9}, {322,9}, {258,9}, {258,9}, {577,9}, {577,9}, {513,9}, {513,9}, {4100,8}, {4100,8}, {4100,8}, {4100,8}, {4545,8}, {4545,8}, {4545,8}, {4545,8}, {4609,8}, {4609,8}, {4609,8}, {4609,8}, {4929,8}, {4929,8}, {4929,8}, {4929,8}, {4865,8}, {4865,8}, {4865,8}, {4865,8}, {4801,8}, {4801,8}, {4801,8}, {4801,8}, {4737,8}, {4737,8}, {4737,8}, {4737,8}, {4673,8}, {4673,8}, {4673,8}, {4673,8}, {10,8}, {10,8}, {10,8}, {10,8}, {385,8}, {385,8}, {385,8}, {385,8}, {194,8}, {194,8}, {194,8}, {194,8}, {68,8}, {68,8}, {68,8}, {68,8}, {449,8}, {449,8}, {449,8}, {449,8}, };
static VLCtabI INTRA_DCT3Dtab2[] = { {5313,11}, {5313,11}, {5377,11}, {5377,11}, {578,11}, {578,11}, {259,11}, {259,11}, {4103,10}, {4103,10}, {4103,10}, {4103,10}, {4163,10}, {4163,10}, {4163,10}, {4163,10}, {4290,10}, {4290,10}, {4290,10}, {4290,10}, {4354,10}, {4354,10}, {4354,10}, {4354,10}, {769,10}, {769,10}, {769,10}, {769,10}, {132,10}, {132,10}, {132,10}, {132,10}, {514,10}, {514,10}, {514,10}, {514,10}, {450,10}, {450,10}, {450,10}, {450,10}, {386,10}, {386,10}, {386,10}, {386,10}, {195,10}, {195,10}, {195,10}, {195,10}, {131,10}, {131,10}, {131,10}, {131,10}, {69,10}, {69,10}, {69,10}, {69,10}, {833,11}, {833,11}, {70,11}, {70,11}, {20,11}, {20,11}, {19,11}, {19,11}, {4227,11}, {4227,11}, {4164,11}, {4164,11}, {4105,11}, {4105,11}, {4104,11}, {4104,11}, {71,12}, {196,12}, {323,12}, {25,12}, {24,12}, {23,12}, {22,12}, {21,12}, {5441,12}, {5505,12}, {5569,12}, {4546,12}, {4482,12}, {4418,12}, {4291,12}, {4106,12}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, {7167,7}, };

void read_H263_getblock(int comp, int mode, int INTRA_AC_DC, int Mode)
{
	int val, i, j = 0, sign;
	unsigned int code;
	VLCtabI *tab;
	short *bp;
	int run, last, level, QP;
	short *qval;


	/* store_code used in alternative inter VLC mode */
	int use_intra = 0;
	CodeCoeff store_code[64];
	int coeff_ind;

	bp = read_H263_ld->block[comp];

	/* decode AC coefficients (or all coefficients in advanced intra coding
	 * mode) */
	if (read_H263_alternative_inter_VLC_mode && !(Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) {
		coeff_ind = 0;
		use_intra = 0;
		for (i = (mode == 0);; i++) {
			code = read_H263_showbits(12);
			store_code[coeff_ind].code = code;
			if (code >= 512) tab = &DCT3Dtab0[(code >> 5) - 16];
			else if (code >= 128) tab = &DCT3Dtab1[(code >> 2) - 32];
			else if (code >= 8) tab = &DCT3Dtab2[(code >> 0) - 8];
			else {
				if (!read_H263_quiet) fprintf (stderr, "invalid Huffman code in getblock()\n");
				read_H263_fault = 1;
				return;
			}

			run = (tab->val >> 4) & 255;
			last = (tab->val >> 12) & 1;
			read_H263_flushbits(tab->len);
			if (tab->val == ESCAPE) {
				last = read_H263_getbits1();
				store_code[coeff_ind].last = last;

				i += run = read_H263_getbits(6);
				store_code[coeff_ind].run = run;

				level = read_H263_getbits(8);
				store_code[coeff_ind].level = level;

				if (level == 128 && read_H263_modified_quantization_mode) {
					level = read_H263_getbits(11);
					store_code[coeff_ind].extended_level = level;
				}
			} else {
				store_code[coeff_ind].sign = read_H263_getbits(1);
				i += run;
			}
			coeff_ind += 1;

			if (i >= 64) use_intra = 1;
			if (last) break;
		}
	}
	coeff_ind = 0;

	for (i = (mode == 0);; i++) {
		if ((read_H263_alternative_inter_VLC_mode) && !(Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) {
			if (use_intra) {
				if ((store_code[coeff_ind].code) >= 512) tab = &INTRA_DCT3Dtab0[((store_code[coeff_ind].code) >> 5) - 16];
				else if ((store_code[coeff_ind].code) >= 128) tab = &INTRA_DCT3Dtab1[((store_code[coeff_ind].code) >> 2) - 32];
				else if ((store_code[coeff_ind].code) >= 8) tab = &INTRA_DCT3Dtab2[((store_code[coeff_ind].code) >> 0) - 8];
				else {
					if (!read_H263_quiet) fprintf (stderr, "invalid Huffman code in getblock()\n");
					read_H263_fault = 1;
					return;
				}
				run = (tab->val >> 6) & 63;
				level = tab->val & 63;
				last = (tab->val >> 12) & 1;
			} else {
				if (store_code[coeff_ind].code >= 512) tab = &DCT3Dtab0[(store_code[coeff_ind].code >> 5) - 16];
				else if (store_code[coeff_ind].code >= 128) tab = &DCT3Dtab1[(store_code[coeff_ind].code >> 2) - 32];
				else if (store_code[coeff_ind].code >= 8) tab = &DCT3Dtab2[(store_code[coeff_ind].code >> 0) - 8];
				else {
					if (!read_H263_quiet) fprintf (stderr, "invalid Huffman code in getblock()\n");
					read_H263_fault = 1;
					return;
				}

				run = (tab->val >> 4) & 255;
				level = tab->val & 15;
				last = (tab->val >> 12) & 1;
			}

			if (tab->val == ESCAPE) {                         /* escape */
				last = store_code[coeff_ind].last;
				i += run = store_code[coeff_ind].run;
				level = store_code[coeff_ind].level;

				if ((sign = (level >= 128))) val = 256 - level;
				else val = level;

				if (level == 128 && read_H263_modified_quantization_mode) {
					level = store_code[coeff_ind].extended_level;
					level = (level >> 6 & 0x001F) | level << 5;
					/* correctly set the sign */
					level = (level << (sizeof (int) * 8 - 11)) >> (sizeof (int) * 8 - 11);
					if (level >= 0) sign = 0;
					else sign = 1;
					val = abs(level);
				}
			} else {
				sign = store_code[coeff_ind].sign;
				i += run;
				val = level;
			}
			coeff_ind += 1;
		} else {
			code = read_H263_showbits(12);
			if ((read_H263_advanced_intra_coding) && (Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) {
				if (code >= 512) tab = &INTRA_DCT3Dtab0[(code >> 5) - 16];
				else if (code >= 128) tab = &INTRA_DCT3Dtab1[(code >> 2) - 32];
				else if (code >= 8) tab = &INTRA_DCT3Dtab2[(code >> 0) - 8];
				else {
					if (!read_H263_quiet) fprintf (stderr, "invalid Huffman code in getblock()\n");
					read_H263_fault = 1;
					return;
				}
				run = (tab->val >> 6) & 63;
				level = tab->val & 63;
				last = (tab->val >> 12) & 1;
			} else {

				if (code >= 512) tab = &DCT3Dtab0[(code >> 5) - 16];
				else if (code >= 128) tab = &DCT3Dtab1[(code >> 2) - 32];
				else if (code >= 8) tab = &DCT3Dtab2[(code >> 0) - 8];
				else {
					if (!read_H263_quiet) fprintf (stderr, "invalid Huffman code in getblock()\n");
					read_H263_fault = 1;
					return;
				}

				run = (tab->val >> 4) & 255;
				level = tab->val & 15;
				last = (tab->val >> 12) & 1;
			}
			read_H263_flushbits(tab->len);
			if (tab->val == ESCAPE) {                         /* escape */
				last = read_H263_getbits1();
				i += run = read_H263_getbits(6);
				level = read_H263_getbits(8);

				if ((sign = (level >= 128))) val = 256 - level;
				else val = level;

				if (level == 128 && read_H263_modified_quantization_mode) {
					/* extended escape code was received */
					level = read_H263_getbits(11);
					level = (level >> 6 & 0x001F) | level << 5;
					/* correctly set the sign */
					level = (level << (sizeof (int) * 8 - 11)) >> (sizeof (int) * 8 - 11);
					if (level >= 0) sign = 0;
					else sign = 1;
					val = abs(level);
				}
			} else {
				i += run;
				val = level;
				sign = read_H263_getbits(1);
			}
		}

		if (i >= 64) {
			if (!read_H263_quiet) fprintf (stderr, "DCT coeff index (i) out of bounds\n");
			read_H263_fault = 1;
			return;
		}

		/* Descan in the proper order in advanced intra coding mode */
		if (read_H263_advanced_intra_coding && (Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) {
			switch (INTRA_AC_DC) {
			case INTRA_MODE_DC:
				j = read_H263_zig_zag_scan[i];
				break;
			case INTRA_MODE_VERT_AC:
				j = read_H263_alternate_horizontal_scan[i];
				break;
			case INTRA_MODE_HORI_AC:
				j = read_H263_alternate_vertical_scan[i];
				break;
			default:
				ERROR(MSG("Invalid Intra_Mode in Advanced Intra Coding"));
				END(1);
				break;
			}
		} else {
			j = read_H263_zig_zag_scan[i];
		}
		qval = &bp[j];
		if (comp >= 6) QP = mmax (1, mmin (31, (read_H263_bquant_tab[read_H263_bquant] * read_H263_quant) >> 2));
		else QP = read_H263_quant;

		if (read_H263_modified_quantization_mode && (comp == 4 || comp == 5 || comp == 10 || comp == 11)) {
			/* when modified quantization mode is on, use modified quantizer for
			 * chorominance coefficients */

			QP = read_H263_MQ_chroma_QP_table[QP];
		}
		if ((read_H263_advanced_intra_coding) && (Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) {
			/* Advanced intra coding dequantization */
			*qval = (sign ? -(QP * (2 * val)) : QP * (2 * val));
		} else {
			/* TMN3 dequantization */
			if ((QP % 2) == 1)
				*qval = (sign ? -(QP * (2 * val + 1)) : QP * (2 * val + 1));
			else
				*qval = (sign ? -(QP * (2 * val + 1) - 1) : QP * (2 * val + 1) - 1);
		}
		if (last) return;
	}
}

/*********************************************************************
 *
 *        Name:        get_sac_block
 *
 *	Description:	Decodes blocks of Arithmetic Encoded DCT Coeffs.
 *        and performs Run Length Decoding and Coefficien
 *        Dequantisation.
 *
 *	Input:        Picture block type and number.
 *
 *	Returns:	Nothing.
 *
 *	Side Effects:
 *
 *	Author:        Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/


void read_H263_get_sac_block (int comp, int ptype, int INTRA_AC_DC, int Mode)
{
	int position = 0;
	int TCOEF_index, symbol_word;
	int last = 0, QP, i, j = 0;
	short *qval;
	short *bp;
	RunCoef DCTcoef;
	int intra;


	bp = read_H263_ld->block[comp];

	i = (ptype == 0);
	intra = (Mode == MODE_INTRA || Mode == MODE_INTRA_Q);

	while (!last) {                             /* while there are DCT coefficients
						     * remaining */
		position++;                 /* coefficient counter relates to Coeff.
					     * model */
		TCOEF_index = read_H263_DecodeTCoef(position, intra);
		if (TCOEF_index == ESCAPE_INDEX) {                           /* ESCAPE code encountered */
			DCTcoef = read_H263_Decode_Escape_Char(intra, &last);
		} else {
			symbol_word = read_H263_tcoeftab[TCOEF_index];
			DCTcoef = read_H263_vlc_word_decode(symbol_word, &last);
		}
		i += DCTcoef.run;
		if (read_H263_advanced_intra_coding && (Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) {
			switch (INTRA_AC_DC) {
			case INTRA_MODE_DC:
				j = read_H263_zig_zag_scan[i];
				break;
			case INTRA_MODE_VERT_AC:
				j = read_H263_alternate_horizontal_scan[i];
				break;
			case INTRA_MODE_HORI_AC:
				j = read_H263_alternate_vertical_scan[i];
				break;
			default:
				ERROR(MSG("Invalid Intra_Mode in Advanced Intra Coding"));
				END(1);
				break;
			}
		} else {
			j = read_H263_zig_zag_scan[i];
		}

		qval = &bp[j];

		i++;

		if (comp >= 6) QP = mmax (1, mmin (31, (read_H263_bquant_tab[read_H263_bquant] * read_H263_quant) >> 2));
		else QP = read_H263_quant;
		if ((read_H263_advanced_intra_coding) && (Mode == MODE_INTRA || Mode == MODE_INTRA_Q)) {
			/* Advanced intra coding dequantization */
			*qval = (DCTcoef.sign ? -(QP * (2 * DCTcoef.val)) : QP * (2 * DCTcoef.val));
		} else {
			if ((QP % 2) == 1) *qval = ((DCTcoef.sign) ? -(QP * (2 * (DCTcoef.val) + 1)) : QP * (2 * (DCTcoef.val) + 1));
			else *qval = ((DCTcoef.sign) ? -(QP * (2 * (DCTcoef.val) + 1) - 1) : QP * (2 * (DCTcoef.val) + 1) - 1);
		}
	}
	return;
}

/*********************************************************************
 *
 *        Name:        vlc_word_decode
 *
 *	Description:	Fills Decoder FIFO after a fixed word length
 *        string has been detected.
 *
 *	Input:        Symbol to be decoded, last data flag.
 *
 *	Returns:	Decoded Symbol via the structure DCTcoeff.
 *
 *	Side Effects:	Updates last flag.
 *
 *	Author:        Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/

RunCoef read_H263_vlc_word_decode(int symbol_word, int *last)
{
	int sign_index;
	RunCoef DCTcoef;

	*last = (symbol_word >> 12) & 01;

	DCTcoef.run = (symbol_word >> 4) & 255;

	DCTcoef.val = (symbol_word) & 15;

	sign_index = read_H263_decode_a_symbol(read_H263_cumf_SIGN);

	DCTcoef.sign = read_H263_signtab[sign_index];

	return (DCTcoef);
}

/*********************************************************************
 *
 *        Name:        Decode_Escape_Char
 *
 *	Description:	Decodes all components for a Symbol when an
 *        ESCAPE character has been detected.
 *
 *	Input:        Picture Type and last data flag.
 *
 *	Returns:	Decoded Symbol via the structure DCTcoeff.
 *
 *	Side Effects:	Modifies last data flag.
 *
 *	Author:        Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/

RunCoef read_H263_Decode_Escape_Char (int intra, int *last)
{
	int last_index, run, run_index, level, level_index;
	RunCoef DCTcoef;

	if (intra) {
		last_index = read_H263_decode_a_symbol(read_H263_cumf_LAST_intra);
		*last = read_H263_last_intratab[last_index];
	} else {
		last_index = read_H263_decode_a_symbol(read_H263_cumf_LAST);
		*last = read_H263_lasttab[last_index];
	}

	if (intra) run_index = read_H263_decode_a_symbol(read_H263_cumf_RUN_intra);
	else run_index = read_H263_decode_a_symbol(read_H263_cumf_RUN);

	run = read_H263_runtab[run_index];

	DCTcoef.run = run;

	if (intra) level_index = read_H263_decode_a_symbol(read_H263_cumf_LEVEL_intra);
	else level_index = read_H263_decode_a_symbol(read_H263_cumf_LEVEL);

	level = read_H263_leveltab[level_index];

	if (level > 128) level -= 256;
	if (level < 0) {
		DCTcoef.sign = 1;
		DCTcoef.val = abs(level);
	} else {
		DCTcoef.sign = 0;
		DCTcoef.val = level;
	}

	return (DCTcoef);
}

/*********************************************************************
 *
 *        Name:        DecodeTCoef
 *
 *	Description:	Decodes a.c DCT Coefficients using the
 *        relevant arithmetic decoding model.
 *
 *	Input:        DCT Coeff count and Picture Type.
 *
 *	Returns:	Index to LU
 *
 *	Side Effects:	None
 *
 *	Author:        Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/
int read_H263_DecodeTCoef (int position, int intra)
{
	int index;

	switch (position) {
	case 1:
		if (intra) index = read_H263_decode_a_symbol(read_H263_cumf_TCOEF1_intra);
		else index = read_H263_decode_a_symbol(read_H263_cumf_TCOEF1);
		break;
	case 2:
		if (intra) index = read_H263_decode_a_symbol(read_H263_cumf_TCOEF2_intra);
		else index = read_H263_decode_a_symbol(read_H263_cumf_TCOEF2);
		break;
	case 3:
		if (intra) index = read_H263_decode_a_symbol(read_H263_cumf_TCOEF3_intra);
		else index = read_H263_decode_a_symbol(read_H263_cumf_TCOEF3);
		break;
	default:
		if (intra) index = read_H263_decode_a_symbol(read_H263_cumf_TCOEFr_intra);
		else index = read_H263_decode_a_symbol(read_H263_cumf_TCOEFr);
		break;
	}

	return (index);
}
