/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

// see below for adding new transport modules

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/time.h>
#define TRANSPORT_INTERNAL
#include "transport.h"
#define ERROR_DEBUG
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) transport_ ## x

int transport_unread_cnt = 0;
int transport_unread_data;
int transport_sunread_cnt = 0;
int transport_sunread_data;
int transport_rformat = 0;
int transport_wformat = 0;
int transport_sync = 0;
int transport_srate = 0;
int transport_rate = 0;
char transport_rname[PATH_MAX] = "./";
char transport_wname[PATH_MAX] = "./";
char transport_rcname[PATH_MAX] = "./";
char transport_wcname[PATH_MAX] = "./";
char transport_srname[PATH_MAX] = "./";
char transport_swname[PATH_MAX] = "./";
char transport_srcname[PATH_MAX] = "./";
char transport_swcname[PATH_MAX] = "./";
char *transport_rdef = NULL, *transport_wdef = NULL;
char *transport_srdef = NULL, *transport_swdef = NULL;
int transport_rfd = -1;
int transport_wfd = -1;
FILE *transport_rfilep = NULL;
FILE *transport_wfilep = NULL;
int transport_srfd = -1;
int transport_swfd = -1;
FILE *transport_srfilep = NULL;
FILE *transport_swfilep = NULL;
int transport_rmulti = 1;
int transport_wmulti = 1;
struct transport_signature transport_rsig, transport_wsig;
struct transport_ssignature transport_srsig, transport_swsig;

 void (*transport_otrace)(const char *s);
 void (*transport_itrace)(const char *s);

/************ Prototypes ************/

void transport_init();
void transport_rdetect(char *format, int *width, int *height);
void transport_srdetect(char *format);
void transport_rmkdef();
void transport_srmkdef();
void transport_swmkdef();
void transport_wmkdef();
int transport_srstart();
int transport_rstart(int multi);
int transport_swstart(const char *format);
int transport_wstart(const char *format, int width, int height, int multi);
void transport_srstop();
void transport_rstop();
void transport_wstop();
void transport_swstop();
void transport_srclose();
void transport_rclose();
void transport_swclose();
void transport_wclose();
void transport_end();
FILE *transport_sfwframe(int nr);
FILE *transport_fwframe(int nr);
FILE *transport_sfwstart(const char *format);
FILE *transport_fwstart(const char *format, int width, int height, int multi);
int transport_wframe(int nr);
int transport_swframe(int nr);
void transport_write(void *ptr, unsigned long cnt);
void transport_swrite(void *ptr, unsigned long cnt);
static inline void transport_write_internal(void *ptr, int cnt, int fd, FILE *fp, const char *name, char *buffer, int *bufpos, int *bits, int bps, struct timeval *begin);
FILE *transport_frframe(int *nr);
FILE *transport_sfrframe(int *nr);
FILE *transport_frstart(int multi);
FILE *transport_sfrstart();
int transport_srframe(int *nr);
int transport_rframe(int *nr);
static inline int transport_tryread_internal(void *ptr, unsigned long cnt, FILE *fp, int fd, const char *name);
int transport_tryread(void *ptr, unsigned long cnt);
int transport_stryread(void *ptr, unsigned long cnt);
void transport_read(void *ptr, unsigned long cnt);
void transport_sread(void *ptr, unsigned long cnt);
void transport_rremove(int nr);
void transport_wremove(int nr);
void transport_write_int(int x);
void transport_swrite_int(int x);
int transport_read_int();
int transport_sread_int();
int transport_tryread_int();
void transport_swlseek(off_t offset, int whence);
void transport_srlseek(off_t offset, int whence);
void transport_unread_int(int n);
void transport_unframe();

/************************************/

#define ADD_FORMAT(symbol) { \
	transport_ ## symbol ## _srlseek, \
	transport_ ## symbol ## _swlseek, \
	transport_ ## symbol ## _srstart, \
	transport_ ## symbol ## _swstart, \
	transport_ ## symbol ## _srframe, \
	transport_ ## symbol ## _swframe, \
	transport_ ## symbol ## _srstop, \
	transport_ ## symbol ## _swstop, \
	transport_ ## symbol ## _srdetect, \
	transport_ ## symbol ## _rstart, \
	transport_ ## symbol ## _wstart, \
	transport_ ## symbol ## _rframe, \
	transport_ ## symbol ## _wframe, \
	transport_ ## symbol ## _rremove, \
	transport_ ## symbol ## _wremove, \
	transport_ ## symbol ## _rstop, \
	transport_ ## symbol ## _wstop, \
	transport_ ## symbol ## _unframe, \
	transport_ ## symbol ## _rdetect, \
	#symbol \
}

#define PROTO(sym) \
 void transport_ ## sym ## _srlseek(off_t offset, int whence); \
 void transport_ ## sym ## _swlseek(off_t offset, int whence); \
 void transport_ ## sym ## _srstart(); \
 void transport_ ## sym ## _swstart(); \
 void transport_ ## sym ## _srframe(int *nr); \
 void transport_ ## sym ## _swframe(int nr); \
 void transport_ ## sym ## _srstop(); \
 void transport_ ## sym ## _swstop(); \
 void transport_ ## sym ## _srdetect(); \
 void transport_ ## sym ## _rstart(); \
 void transport_ ## sym ## _wstart(); \
 void transport_ ## sym ## _rframe(int *nr); \
 void transport_ ## sym ## _wframe(int nr); \
 void transport_ ## sym ## _rremove(int nr); \
 void transport_ ## sym ## _wremove(int nr); \
 void transport_ ## sym ## _rstop(); \
 void transport_ ## sym ## _wstop(); \
 void transport_ ## sym ## _unframe(); \
 void transport_ ## sym ## _rdetect();

struct TRANSPORT_FORMAT_STRUCT {
	void (*srlseek)(off_t offset, int whence);
	void (*swlseek)(off_t offset, int whence);
	void (*srstart)();
	void (*swstart)();
	void (*srframe)(int *nr);
	void (*swframe)(int nr);
	void (*srstop)();
	void (*swstop)();
	void (*srdetect)();
	void (*rstart)();
	void (*wstart)();
	void (*rframe)(int *nr);
	void (*wframe)(int nr);
	void (*rremove)(int nr);
	void (*wremove)(int nr);
	void (*rstop)();
	void (*wstop)();
	void (*unframe)();
	void (*rdetect)();
	char *name;
};

////////////////////////////////////////////////////////////////////////////////
///////////////////////// transport module registration ////////////////////////
////////////////////////////////////////////////////////////////////////////////

// Step 1: define prototypes
PROTO(file)
PROTO(net)
PROTO(webcam)
PROTO(mc)

// Step 2: increase number of formats
#define TRANSPORT_FORMATS 4

// Step 3: fill in structure with ADD_FORMAT
//         supply symbol name
struct TRANSPORT_FORMAT_STRUCT transport_formats[TRANSPORT_FORMATS] = {
	ADD_FORMAT(file),
	ADD_FORMAT(webcam),
	ADD_FORMAT(mc),
	ADD_FORMAT(net)
};

// Step 4: add the new object file to Makefile.in
// Step 5: create the source and header file, see transport_file.c/h for an example

// finished
////////////////////////////////////////////////////////////////////////////////

START_OPTIONS
STROPTION(rname);
STROPTION(wname);
STROPTION(srname);
STROPTION(swname);
INTOPTION(rformat,0,TRANSPORT_FORMATS-1);
INTOPTION(wformat,0,TRANSPORT_FORMATS-1);
OPTION(format nr) NOSET O_GET sprintf(value,"%i",TRANSPORT_FORMATS);
OPTION(rformat str) O_SET {
	int i;
	for (i = 0; i < TRANSPORT_FORMATS; i++) {
		if (!strcmp(transport_formats[i].name,value)) {
			transport_rformat = i;
			break;
		}
	}
	ASSERT(i < TRANSPORT_FORMATS,(MSG("unknown transport format '%s'"),value));
} O_GET {
	strcpy(value,transport_formats[transport_rformat].name);
}
OPTION(wformat str) O_SET {
	int i;
	for (i = 0; i < TRANSPORT_FORMATS; i++) {
		if (!strcmp(transport_formats[i].name,value)) {
			transport_wformat = i;
			break;
		}
	}
	ASSERT(i < TRANSPORT_FORMATS,(MSG("unknown transport format '%s'"),value));
} O_GET {
	strcpy(value,transport_formats[transport_wformat].name);
}
INTOPTION(sync,0,1);
INTOPTION(rate,0,INT_MAX);
INTOPTION(srate,0,INT_MAX);
END_OPTIONS

void transport_init()
{
}

void transport_rdetect(char *format, int *width, int *height)
{
	transport_rsig.version = TRANSPORT_CURRENT_VERSION;
	transport_rsig.width = *width;
	transport_rsig.height = *height;
	strcpy(transport_rsig.format,format);

	(*transport_formats[transport_rformat].rdetect)();

	ASSERT(transport_rsig.version == TRANSPORT_CURRENT_VERSION,(MSG("input stream is rev. %i, need rev. %i"),transport_rsig.version,TRANSPORT_CURRENT_VERSION));
	*width = transport_rsig.width;
	*height = transport_rsig.height;
	strcpy(format,transport_rsig.format);
}

void transport_srdetect(char *format)
{
	transport_srsig.version = TRANSPORT_CURRENT_VERSION;
	strcpy(transport_srsig.format,format);

	(*transport_formats[transport_rformat].srdetect)();

	ASSERT(transport_srsig.version == TRANSPORT_CURRENT_VERSION,(MSG("input stream is rev. %i, need rev. %i"),transport_rsig.version,TRANSPORT_CURRENT_VERSION));
	strcpy(format,transport_srsig.format);
}

void transport_rmkdef()
{
	transport_rdef = transport_rname+strlen(transport_rname);
	if (transport_rmulti) {
		strcat(transport_rname,"p_%04d.");
	} else {
		strcat(transport_rname,"output.");
	}
	strcat(transport_rname,transport_rsig.format);
}

void transport_srmkdef()
{
	transport_srdef = transport_srname+strlen(transport_srname);
	strcat(transport_srname,"output.");
	strcat(transport_srname,transport_srsig.format);
}

void transport_swmkdef()
{
	transport_swdef = transport_swname+strlen(transport_swname);
	strcat(transport_swname,"output.");
	strcat(transport_swname,transport_swsig.format);
}

void transport_wmkdef()
{
	transport_wdef = transport_wname+strlen(transport_wname);
	if (transport_wmulti) {
		strcat(transport_wname,"p_%04d.");
	} else {
		strcat(transport_wname,"output.");
	}
	strcat(transport_wname,transport_wsig.format);
}

int transport_srstart()
{
	(*transport_formats[transport_rformat].srstart)();
	return transport_srfd;
}

int transport_rstart(int multi)
{
	transport_rmulti = multi;
	transport_unread_cnt = 0;
	(*transport_formats[transport_rformat].rstart)();
	return transport_rfd;
}

int transport_swstart(const char *format)
{
	transport_swsig.version = TRANSPORT_CURRENT_VERSION;
	strcpy(transport_swsig.format,format);
	(*transport_formats[transport_wformat].swstart)();

	return transport_swfd;
}

int transport_wstart(const char *format, int width, int height, int multi)
{
	(*transport_otrace)(transport_wname);
	transport_wmulti = multi;
	transport_wsig.version = TRANSPORT_CURRENT_VERSION;
	transport_wsig.width = width;
	transport_wsig.height = height;
	strcpy(transport_wsig.format,format);
	(*transport_formats[transport_wformat].wstart)();

	return transport_wfd;
}

void transport_srstop()
{
	(*transport_formats[transport_rformat].srstop)();
	transport_srclose();
	if (transport_srdef) *transport_srdef = 0;
	transport_srdef = NULL;
}

void transport_rstop()
{
	(*transport_formats[transport_rformat].rstop)();
	transport_rclose();
	if (transport_rdef) *transport_rdef = 0;
	transport_rdef = NULL;
}

void transport_wstop()
{
	(*transport_formats[transport_wformat].wstop)();
	transport_wclose();
	if (transport_wdef) *transport_wdef = 0;
	transport_wdef = NULL;
}

void transport_swstop()
{
	(*transport_formats[transport_wformat].swstop)();
	transport_swclose();
	if (transport_swdef) *transport_swdef = 0;
	transport_swdef = NULL;
}

void transport_srclose()
{
	if (transport_srfd >= 0) {
		if (transport_srfilep) {
			SAFE_FCLOSE(transport_srfilep,transport_srcname);
			transport_srfilep = NULL;
		} else {
			SAFE_CLOSE(transport_srfd,transport_srcname);
		}
		transport_srfd = -1;
	}
}

void transport_rclose()
{
	if (transport_rfd >= 0) {
		if (transport_rfilep) {
			SAFE_FCLOSE(transport_rfilep,transport_rcname);
			transport_rfilep = NULL;
		} else {
			SAFE_CLOSE(transport_rfd,transport_rcname);
		}
		transport_rfd = -1;
	}
}

void transport_swclose()
{
	if (transport_swfd >= 0) {
		transport_swrite(NULL,0);
		if (transport_swfilep) {
			SAFE_FCLOSE(transport_swfilep,transport_swcname);
			transport_swfilep = NULL;
		} else {
			SAFE_CLOSE(transport_swfd,transport_swcname);
		}
		transport_swfd = -1;
	}
}

void transport_wclose()
{
	if (transport_wfd >= 0) {
		transport_write(NULL,0);
		if (transport_wfilep) {
			SAFE_FCLOSE(transport_wfilep,transport_wcname);
			transport_wfilep = NULL;
		} else {
			SAFE_CLOSE(transport_wfd,transport_wcname);
		}
		transport_wfd = -1;
	}
}

void transport_end()
{
	transport_rstop();
	transport_wstop();
	transport_srstop();
	transport_swstop();
}

FILE *transport_sfwframe(int nr)
{
	transport_swframe(nr);
	if (!transport_swfilep) NOTNULL((transport_swfilep = fdopen(transport_swfd,"wb")),OPENMSG(transport_swcname));
	return transport_swfilep;
}

FILE *transport_fwframe(int nr)
{
	transport_wframe(nr);
	if (!transport_wfilep) NOTNULL((transport_wfilep = fdopen(transport_wfd,"wb")),OPENMSG(transport_wcname));
	return transport_wfilep;
}

FILE *transport_sfwstart(const char *format)
{
	transport_swstart(format);
	if (!transport_swfilep && transport_swfd != -1) NOTNULL((transport_swfilep = fdopen(transport_swfd,"wb")),OPENMSG(transport_swname));
	return transport_swfilep;
}

FILE *transport_fwstart(const char *format, int width, int height, int multi)
{
	transport_wstart(format, width, height, multi);
	if (!transport_wfilep && transport_wfd != -1) NOTNULL((transport_wfilep = fdopen(transport_wfd,"wb")),OPENMSG(transport_wname));
	return transport_wfilep;
}

int transport_wframe(int nr)
{
	char tmp[PATH_MAX];
	transport_write(NULL,0);
	transport_swrite(NULL,0);

#ifdef HAVE_FDATASYNC
	if (transport_sync) fdatasync(transport_wfd);
#else
	if (transport_sync) fsync(transport_wfd);
#endif
	if (transport_wmulti) {
		transport_wclose();
		sprintf(transport_wcname, transport_wname, nr);
		sprintf(tmp,"%s [%i]",transport_wcname,nr);
	} else {
		sprintf(tmp,"%s [%i]",transport_wname,nr);
	}
	(*transport_otrace)(tmp);

	(*transport_formats[transport_wformat].wframe)(nr);

	return transport_wfd;
}

int transport_swframe(int nr)
{
	transport_write(NULL,0);
	transport_swrite(NULL,0);

#ifdef HAVE_FDATASYNC
	if (transport_sync) fdatasync(transport_swfd);
#else
	if (transport_sync) fsync(transport_swfd);
#endif
	(*transport_formats[transport_wformat].swframe)(nr);

	return transport_swfd;
}

#define THRESHOLD 64
#define BUFSIZE 128

void transport_write(void *ptr, unsigned long cnt)
{
	static char buffer[BUFSIZE+THRESHOLD];
	static int bufpos = 0;
	static int bits = 0;
	static struct timeval begin = {0,0};
	if (begin.tv_sec == 0 && begin.tv_usec == 0) gettimeofday(&begin,0);
	transport_write_internal(ptr, cnt, transport_wfd, transport_wfilep, transport_wcname, buffer, &bufpos, &bits, transport_rate, &begin);
}

void transport_swrite(void *ptr, unsigned long cnt)
{
	static char buffer[BUFSIZE+THRESHOLD];
	static int bufpos = 0;
	static int bits = 0;
	static struct timeval begin = {0,0};
	if (begin.tv_sec == 0 && begin.tv_usec == 0) gettimeofday(&begin,0);
	transport_write_internal(ptr, cnt, transport_swfd, transport_swfilep, transport_swcname, buffer, &bufpos, &bits, transport_srate, &begin);
}

static inline void transport_write_internal(void *ptr, int cnt, int fd, FILE *fp, const char *name, char *buffer, int *bufpos, int *bits, int bps, struct timeval *begin)
{
	int x;
	if (cnt > 0) {
		if (cnt <= THRESHOLD) {
			memcpy(buffer+*bufpos,ptr,cnt);
			*bufpos += cnt;
			cnt = 0;
			if (*bufpos < BUFSIZE) return;
		}
		if (*bufpos) transport_write_internal(buffer,0,fd,fp,name,buffer,bufpos,bits,bps,begin);
	} else {
		cnt = *bufpos;
		ptr = buffer;
		*bufpos = 0;
	}
	if (bps) {
		struct timeval now;
		int time;
		gettimeofday(&now,0);
		time = ((now.tv_sec*1000)+(now.tv_usec/1000))
			-((begin->tv_sec*1000)+(begin->tv_usec/1000));
		*bits -= (time*bps)/1000;
		if (*bits > 0) {
			usleep((((*bits)*1000)/bps)*1000);
		}
		*bits = cnt*8;
		gettimeofday(begin,0);
	}
	if (fp) {
		while (cnt > 0) {
			ASSERT((x = fwrite(ptr,01,cnt,fp)) > 0,WRITEMSG(name));
			if (x == 0) return;
			cnt -= x;
			ptr += x;
		}
		if (*bufpos == 0) fflush(fp);
	} else {
		while (cnt > 0) {
			ASSERT((x = write(fd,ptr,cnt)) >= 0,WRITEMSG(name));
			if (x == -1) return;
			cnt -= x;
			ptr += x;
		}
	}
}

FILE *transport_frframe(int *nr)
{
	transport_rframe(nr);
	if (!transport_rfilep) NOTNULL((transport_rfilep = fdopen(transport_rfd,"rb")),OPENMSG(transport_rcname));
	return transport_rfilep;
}

FILE *transport_sfrframe(int *nr)
{
	transport_srframe(nr);
	if (!transport_srfilep) NOTNULL((transport_srfilep = fdopen(transport_srfd,"rb")),OPENMSG(transport_srcname));
	return transport_srfilep;
}

FILE *transport_frstart(int multi)
{
	transport_rstart(multi);
	if (!transport_rfilep && transport_rfd != -1) NOTNULL((transport_rfilep = fdopen(transport_rfd,"rb")),OPENMSG(transport_rname));
	return transport_rfilep;
}

FILE *transport_sfrstart()
{
	transport_srstart();
	if (!transport_srfilep && transport_srfd != -1) NOTNULL((transport_srfilep = fdopen(transport_srfd,"rb")),OPENMSG(transport_srname));
	return transport_srfilep;
}

int transport_srframe(int *nr)
{
	(*transport_formats[transport_rformat].srframe)(nr);

	return transport_srfd;
}

int transport_rframe(int *nr)
{
	char tmp[PATH_MAX];

	if (transport_rmulti) {
		transport_rclose();
		sprintf(transport_rcname, transport_rname,*nr);
		sprintf(tmp,"%s [%i]",transport_rcname,*nr);
	} else {
		sprintf(tmp,"%s [%i]",transport_rname,*nr);
	}
	(*transport_itrace)(tmp);

	(*transport_formats[transport_rformat].rframe)(nr);

	return transport_rfd;
}

#define RBUFSIZE 65536
static inline int transport_tryread_internal(void *ptr, unsigned long cnt, FILE *fp, int fd, const char *name)
{
	int x;
	void *start = ptr;

	if (fp) {	
		while (cnt > 0) {
			x = fread(ptr,01,cnt,fp);
#if 0 //def _WIN32
			if (x <= 0 && errno == EAGAIN) Sleep(10);
#endif
			if (errno != EAGAIN && x <= 0) {
				if (ferror(fp)) { ERROR READMSG(name); END(1); }
				return (int)(ptr-start);
			}
			cnt -= x;
			ptr += x;
		}
	} else {
		while (cnt > 0) {
			errno = 0;
			ASSERT((x = read(fd,ptr,cnt)) >= 0,READMSG(name));
#if 0 //def _WIN32
			if (errno == EAGAIN) Sleep(10);
#endif
			if (errno != EAGAIN && (x == 0 || x == -1)) return (int)(ptr-start);
			cnt -= x;
			ptr += x;
		}
	}
	return (int)(ptr-start);
}

int transport_tryread(void *ptr, unsigned long cnt)
{
	return transport_tryread_internal(ptr,cnt,transport_rfilep,transport_rfd,transport_rcname);
}

int transport_stryread(void *ptr, unsigned long cnt)
{
	return transport_tryread_internal(ptr,cnt,transport_srfilep,transport_srfd,transport_srcname);
}

void transport_read(void *ptr, unsigned long cnt)
{
	ASSERT(cnt == transport_tryread(ptr,cnt),(MSG("end of file (read %s)"),transport_rcname));
}

void transport_sread(void *ptr, unsigned long cnt)
{
	ASSERT(cnt == transport_stryread(ptr,cnt),(MSG("end of file (read %s)"),transport_srcname));
}

void transport_rremove(int nr)
{
	if (transport_rmulti) (*transport_formats[transport_rformat].rremove)(nr);
}

void transport_wremove(int nr)
{
	if (transport_wmulti) (*transport_formats[transport_wformat].wremove)(nr);
}

void transport_write_int(int x)
{
	unsigned char c;

	c = x & 0xff;
	transport_write(&c,1);
	x >>= 8;
	c = x & 0xff;
	transport_write(&c,1);
	x >>= 8;
	c = x & 0xff;
	transport_write(&c,1);
	x >>= 8;
	c = x & 0xff;
	transport_write(&c,1);
}

void transport_swrite_int(int x)
{
	unsigned char c;

	c = x & 0xff;
	transport_swrite(&c,1);
	x >>= 8;
	c = x & 0xff;
	transport_swrite(&c,1);
	x >>= 8;
	c = x & 0xff;
	transport_swrite(&c,1);
	x >>= 8;
	c = x & 0xff;
	transport_swrite(&c,1);
}

int transport_read_int()
{
	unsigned char c;
	int x;

	if (transport_unread_cnt) {
		transport_unread_cnt = 0;
		return transport_unread_data;
	}

	transport_read(&c,1);
	x = c;
	transport_read(&c,1);
	x |= c << 8;
	transport_read(&c,1);
	x |= c << 16;
	transport_read(&c,1);
	x |= c << 24;

	return x;
}

int transport_sread_int()
{
	unsigned char c;
	int x;

	if (transport_sunread_cnt) {
		transport_sunread_cnt = 0;
		return transport_sunread_data;
	}

	transport_sread(&c,1);
	x = c;
	transport_sread(&c,1);
	x |= c << 8;
	transport_sread(&c,1);
	x |= c << 16;
	transport_sread(&c,1);
	x |= c << 24;

	return x;
}

int transport_tryread_int()
{
	unsigned char c;
	int x;

	if (transport_unread_cnt) {
		transport_unread_cnt = 0;
		return transport_unread_data;
	}

	if (!transport_tryread(&c,1)) return -1;
	x = c;
	if (!transport_tryread(&c,1)) return -1;
	x |= c << 8;
	if (!transport_tryread(&c,1)) return -1;
	x |= c << 16;
	if (!transport_tryread(&c,1)) return -1;
	x |= c << 24;

	return x;
}

void transport_swlseek(off_t offset, int whence)
{
	(*transport_formats[transport_wformat].swlseek)(offset,whence);
}

void transport_srlseek(off_t offset, int whence)
{
	(*transport_formats[transport_rformat].srlseek)(offset,whence);
}

void transport_unread_int(int n)
{
	transport_unread_cnt = 1;
	transport_unread_data = n;
}

void transport_unframe()
{
	(*transport_formats[transport_rformat].unframe)();
}
