/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <signal.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#define TRANSPORT_INTERNAL
#include "transport.h"
#define ERROR_DEBUG
#include "error.h"

/************ Prototypes ************/

void transport_mc_srlseek(off_t offset, int whence);
void transport_mc_swlseek(off_t offset, int whence);
void transport_mc_wstart();
void transport_mc_swstart();
static inline int transport_mc_iwstart(char *name, int port);
void transport_mc_wframe(int nr);
void transport_mc_swframe(int nr);
void transport_mc_rframe(int *nr);
void transport_mc_srframe(int nr);
void transport_mc_wstop();
void transport_mc_swstop();
void transport_mc_abort(int x);
void transport_mc_rremove(int n);
void transport_mc_wremove(int n);
void transport_mc_rdetect();
void transport_mc_srdetect();
void transport_mc_rstart();
void *transport_mc_keepalive_thread(void *ign);
void transport_mc_srstart();
static inline int transport_mc_irstart(char *name, int port);
void transport_mc_rstop();
void transport_mc_srstop();
void transport_mc_unframe();

/************************************/

int transport_mc_unframe_num = 0;
int transport_mc_unframe_p = 0;
int transport_mc_dwfd = -1;

void transport_mc_srlseek(off_t offset, int whence)
{
	if (whence == SEEK_CUR && offset > 0) {
		char buffer[1024];
		while (offset > 1024) {
			transport_sread(buffer,1024);
			offset -= 1024;
		}
		transport_sread(buffer,offset);
	}
}

void transport_mc_swlseek(off_t offset, int whence)
{
	if (whence == SEEK_CUR && offset > 0) {
		char buffer[1024];
		memset(buffer,0,1024);
		while (offset > 1024) {
			transport_swrite(buffer,1024);
			offset -= 1024;
		}
		transport_swrite(buffer,offset);
	}
}

void transport_mc_wstart()
{
	char tmp[PATH_MAX], *p;
	transport_wmulti = 0;
	transport_wfd = transport_mc_iwstart(transport_wname,31337);
	strcpy(tmp,transport_wname);
	p = strchr(tmp,':');
	if (p) *p = 0;
	transport_mc_dwfd = transport_mc_iwstart(tmp,31336);
	fcntl(transport_mc_dwfd,F_SETFL,O_NONBLOCK);
}

void transport_mc_swstart()
{
	transport_swfd = transport_mc_iwstart(transport_swname,31338);
}

int transport_mc_owstart(char *name,int defport)
{
	return transport_mc_iwstart(name,defport);
}

static inline int transport_mc_iwstart(char *name, int port)
{
	int fd, tmp;
	char *portp;
	struct sockaddr_in net_addr;
	
	portp = strchr(name,':');
	if (portp) {
		*portp = 0;
		port = atoi(portp+1);
		ASSERT(port > 0 && port < 65536,(MSG("invalid port address: %i"),port));
	}

	ASSERT((fd=socket(AF_INET,SOCK_DGRAM,0)) >= 0,(ERRMSG("socket")));
	
	memset(&net_addr,0,sizeof(net_addr));
	net_addr.sin_family=AF_INET;
//	net_addr.sin_addr.s_addr = htonl(INADDR_ANY);
//	net_addr.sin_port=htons(0);
//	bind(fd,(struct sockaddr *)&net_addr,sizeof(net_addr));

	if (name[0] == 0 || (name[0] == '.' && name[1] == 0)) NOTNEG1(net_addr.sin_addr.s_addr = inet_addr("224.0.1.20"),(INTMSG("inet_aton")));
	else NOTNEG1(net_addr.sin_addr.s_addr = inet_addr(name),(MSG("invalid internet address: %s"),name));
	net_addr.sin_port=htons(port);

	ASSERT(connect(fd,(struct sockaddr *)&net_addr,sizeof(net_addr)) == 0,(ERRMSG("connect")));
	tmp = 255;
	ASSERT(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL,(void *)&tmp, sizeof(tmp)) == 0,(ERRMSG("setsockopt")));
	ASSERT(signal(SIGPIPE,transport_mc_abort) != SIG_ERR,(ERRMSG("signal")));

	return fd;
}

void transport_mc_wframe(int nr)
{
	char tmp[80];
	if (transport_wmulti) transport_write_int(nr);
	read(transport_mc_dwfd,tmp,80);
}

void transport_mc_swframe(int nr)
{
}

void transport_mc_rframe(int *nr)
{
	if (transport_rmulti) {
		if (transport_mc_unframe_p) {
			transport_mc_unframe_p = 0;
			*nr = transport_mc_unframe_num;
			return;
		}

		*nr = transport_mc_unframe_num = transport_read_int();
	}
}

void transport_mc_srframe(int nr)
{
}

void transport_mc_wstop()
{
	if (transport_mc_dwfd >= 0) {
		close(transport_mc_dwfd);
		transport_mc_dwfd = -1;
	}
}

void transport_mc_swstop()
{
}

void transport_mc_abort(int x)
{
	ERROR(MSG("Network error (%i)"),x);
	END(x);
}

void transport_mc_rremove(int n)
{
}

void transport_mc_wremove(int n)
{
}


void transport_mc_rdetect()
{
}

void transport_mc_srdetect()
{
}

void transport_mc_rstart()
{
	pthread_attr_t pattr;
	pthread_t thread;

	transport_rmulti = 0;
	transport_rfd = transport_mc_irstart(transport_rname,31337);

	pthread_attr_init(&pattr);
#ifdef PTHREAD_CREATE_DETACHED
	pthread_attr_setdetachstate(&pattr,PTHREAD_CREATE_DETACHED);
#endif
	EQNULL((void *)pthread_create(&thread,&pattr,transport_mc_keepalive_thread,NULL),(ERRMSG("pthread_create")));
	pthread_attr_destroy(&pattr);
}

void *transport_mc_keepalive_thread(void *ign)
{
	char tmp[PATH_MAX>512?PATH_MAX:512], *p;
	fd_set rset, eset;
	int fd, port = 31336;
	struct timeval timeout;

	strcpy(tmp,transport_rname);
	p = strchr(tmp,':');
	if (p) {
		port = atoi(p+1)-1;
		*p = 0;
	}
	fd = transport_mc_irstart(tmp,port);
	memset(tmp,0,512);
	fcntl(fd,F_SETFL,O_NONBLOCK);

	FD_ZERO(&rset);
	FD_ZERO(&eset);
loop:
	timeout.tv_sec = 10;
	timeout.tv_usec = 0;
	FD_SET(transport_rfd,&rset);
	FD_SET(transport_rfd,&eset);

	if (!select(transport_rfd,&rset,NULL,&eset,&timeout)) {
		write(fd,tmp,14);
	} else {
		if (FD_ISSET(transport_rfd,&eset)) {
			close(fd);
			return NULL;
		}
		sleep(60);
	}

	goto loop;
}

void transport_mc_srstart()
{
	transport_srfd = transport_mc_irstart(transport_srname,31338);
}

int transport_mc_orstart(char *name,int defport)
{
	int fd = transport_mc_irstart(name,defport);
	fcntl(fd,F_SETFL,O_NONBLOCK);
	return fd;
}

static inline int transport_mc_irstart(char *name, int port)
{
	char *portp, *remotep;
	int fd, tmp;
	struct sockaddr_in net_addr;
	struct ip_mreq mreq;

	ASSERT((fd=socket(AF_INET,SOCK_DGRAM,0)) >= 0,(ERRMSG("socket")));
	
	memset((void *)&net_addr,0,sizeof(net_addr));
	net_addr.sin_family=AF_INET;
	net_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	net_addr.sin_port=htons(port);

	remotep = strchr(name,'/');
	if (remotep) {
		portp = strchr(remotep+1,':');
		if (portp) {
			*portp = 0;
			ASSERT(atoi(portp+1) >= 0 && atoi(portp+1) < 65536,(MSG("invalid port address: %i"),port));
			net_addr.sin_port=htons(atoi(portp+1));
		}
		if (!(remotep[1] == 0 || (remotep[1] == '.' && remotep[2] == 0))) {
			NOTNEG1(net_addr.sin_addr.s_addr = inet_addr(remotep+1),(MSG("invalid internet address: %s"),remotep+1));
		}
		if (portp) *portp = ':';
		*remotep = 0;
	}
//	ASSERT(connect(fd,(struct sockaddr *)&net_addr,sizeof(net_addr)) >= 0,(ERRMSG("connect")));

	portp = strchr(name,':');
	if (portp) {
		*portp = 0;
		port = atoi(portp+1);
		ASSERT(port > 0 && port < 65536,(MSG("invalid port address: %i"),port));
	}
	if (name[0] == 0 || (name[0] == '.' && name[1] == 0)) {
		NOTNEG1(mreq.imr_multiaddr.s_addr = inet_addr("224.0.1.20"),(INTMSG("inet_aton")));
	} else {
		NOTNEG1(mreq.imr_multiaddr.s_addr = inet_addr(name),(MSG("invalid internet address: %s"),name));
	}
	if (remotep) *remotep = '/';

	memset(&net_addr,0,sizeof(net_addr));
	net_addr.sin_family=AF_INET;
	net_addr.sin_port=htons(port);
	net_addr.sin_addr.s_addr = mreq.imr_multiaddr.s_addr;
	if (bind(fd,(struct sockaddr *)&net_addr,sizeof(net_addr)) < 0) {
		net_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		EQNULL((void *)bind(fd,(struct sockaddr *)&net_addr,sizeof(net_addr)),(ERRMSG("bind")));
	}

	mreq.imr_interface.s_addr=htonl(INADDR_ANY);
	ASSERT(setsockopt(fd,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)) >= 0,(ERRMSG("setsockopt")));

//	tmp = 8192;
//	ASSERT(setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (void *)&tmp, sizeof(tmp)),(ERRMSG("setsockopt")));

	if (portp) *portp = ':';
	else sprintf(name+strlen(name),":%i",port);

#if 0 //def _WIN32
	fcntl(fd,F_SETFL,O_NONBLOCK);
#endif

	return fd;
}

void transport_mc_rstop()
{
}

void transport_mc_srstop()
{
}

void transport_mc_unframe()
{
	transport_mc_unframe_p = 1;
}
